"""Module that defines built-in callback functions for configuration item values."""


import os
import re
import fsspec
from hpcflow.sdk.core.errors import UnsupportedSchedulerError, UnsupportedShellError

from hpcflow.sdk.submission.shells import get_supported_shells


def callback_vars(config, value):
    """
    Callback that substitutes configuration variables.
    """

    def vars_repl(match_obj):
        var_name = match_obj.groups()[0]
        return config._variables[var_name]

    vars_join = "|".join(list(config._variables.keys()))
    vars_regex = r"\<\<(" + vars_join + r")\>\>"
    value = re.sub(
        pattern=vars_regex,
        repl=vars_repl,
        string=str(value),
    )
    return value


def callback_file_paths(config, file_path):
    """
    Callback that resolves file paths.
    """
    if isinstance(file_path, list):
        return [config._resolve_path(i) for i in file_path]
    else:
        return config._resolve_path(file_path)


def callback_bool(config, value):
    """
    Callback that coerces values to boolean.
    """
    if not isinstance(value, bool):
        if value.lower() == "true":
            return True
        elif value.lower() == "false":
            return False
        else:
            raise TypeError(f"Cannot cast {value!r} to a bool type.")
    return value


def callback_lowercase(config, value):
    """
    Callback that forces a string to lower case.
    """
    if isinstance(value, list):
        return [i.lower() for i in value]
    elif isinstance(value, dict):
        return {k.lower(): v for k, v in value.items()}
    else:
        return value.lower()


def exists_in_schedulers(config, value):
    """
    Callback that tests that a value is a supported scheduler name.
    """
    if value not in config.schedulers:
        raise ValueError(
            f"Cannot set default scheduler; {value!r} is not a supported scheduler "
            f"according to the config file, which lists these schedulers as available "
            f"on this machine: {config.schedulers!r}."
        )
    return value


def callback_supported_schedulers(config, schedulers):
    """
    Callback that tests that all values are names of supported schedulers.
    """
    # validate against supported schedulers according to the OS - this won't validate that
    # a particular scheduler actually exists on this system:
    available = config._app.get_OS_supported_schedulers()
    for k in schedulers:
        if k not in available:
            raise UnsupportedSchedulerError(scheduler=k, available=available)

    return schedulers


def set_scheduler_invocation_match(config, scheduler: str):
    """Invoked on set of `default_scheduler`.

    For clusters with "proper" schedulers (SGE, SLURM, etc.), login nodes are typically
    named using the word "login". So we can use this knowledge to set a default for the
    "hostname" invocation match key, if it is not manually set. However, it is preferable
    that on clusters the hostname match is explicitly set.

    """
    default_args = config.get(f"schedulers.{scheduler}").get("defaults", {})
    sched = config._app.get_scheduler(
        scheduler_name=scheduler,
        os_name=os.name,
        scheduler_args=default_args,
    )
    if hasattr(sched, "DEFAULT_LOGIN_NODE_MATCH"):
        if "hostname" not in config._file.get_invocation(config._config_key)["match"]:
            config._file.update_invocation(
                config_key=config._config_key,
                match={"hostname": sched.DEFAULT_LOGIN_NODE_MATCH},
            )


def callback_scheduler_set_up(config, schedulers):
    """Invoked on set of `schedulers`.

    Runs scheduler-specific config initialisation.

    """
    for k, v in schedulers.items():
        sched = config._app.get_scheduler(
            scheduler_name=k,
            os_name=os.name,
            scheduler_args=v.get("defaults", {}),
        )
        if hasattr(sched, "get_login_nodes"):
            # some `Scheduler` classes have a `get_login_nodes` method which can be used
            # to populate the names of login nodes explicitly, if not already set:
            if "hostname" not in config._file.get_invocation(config._config_key)["match"]:
                login_nodes = sched.get_login_nodes()
                config._file.update_invocation(
                    config_key=config._config_key,
                    match={"hostname": login_nodes},
                )
    return schedulers


def callback_supported_shells(config, shell_name):
    """
    Callback that tests if a shell names is supported on this OS.
    """
    supported = get_supported_shells(os.name)
    if shell_name not in supported:
        raise UnsupportedShellError(shell=shell_name, supported=supported)
    return shell_name


def set_callback_file_paths(config, value):
    """Check the file(s) is/are accessible. This is only done on `config.set` (and not on
    `config.get` or `config._validate`) because it could be expensive in the case of remote
    files."""
    value = callback_file_paths(config, value)

    to_check = value
    if not isinstance(value, list):
        to_check = [value]

    for file_path in to_check:
        with fsspec.open(file_path, mode="rt") as fh:
            pass
            # TODO: also check something in it?
        print(f"Checked access to: {file_path}")


def check_load_data_files(config, value):
    """Check data files (e.g., task schema files) can be loaded successfully. This is only
    done on `config.set` (and not on `config.get` or `config._validate`) because it could
    be expensive in the case of remote files."""
    config._app.reload_template_components(warn=False)


def callback_update_log_console_level(config, value):
    """
    Callback to set the logging level.
    """
    config._app.log.update_console_level(value)



"""
A collection of submissions to a scheduler, generated from a workflow.
"""

from __future__ import annotations
from collections import defaultdict

from datetime import datetime, timedelta, timezone
import enum
import os
from pathlib import Path
from textwrap import indent
from typing import Dict, List, Optional, Tuple

from hpcflow.sdk import app
from hpcflow.sdk.core.element import ElementResources
from hpcflow.sdk.core.errors import (
    JobscriptSubmissionFailure,
    MissingEnvironmentError,
    MissingEnvironmentExecutableError,
    MissingEnvironmentExecutableInstanceError,
    MultipleEnvironmentsError,
    SubmissionFailure,
)
from hpcflow.sdk.core.json_like import ChildObjectSpec, JSONLike
from hpcflow.sdk.core.object_list import ObjectListMultipleMatchError
from hpcflow.sdk.log import TimeIt


def timedelta_format(td: timedelta) -> str:
    """
    Convert time delta to string in standard form.
    """
    days, seconds = td.days, td.seconds
    hours = seconds // (60 * 60)
    seconds -= hours * (60 * 60)
    minutes = seconds // 60
    seconds -= minutes * 60
    return f"{days}-{hours:02}:{minutes:02}:{seconds:02}"


def timedelta_parse(td_str: str) -> timedelta:
    """
    Parse a string in standard form as a time delta.
    """
    days, other = td_str.split("-")
    days = int(days)
    hours, mins, secs = [int(i) for i in other.split(":")]
    return timedelta(days=days, hours=hours, minutes=mins, seconds=secs)


class SubmissionStatus(enum.Enum):
    """
    The overall status of a submission.
    """

    #: Not yet submitted.
    PENDING = 0
    #: All jobscripts submitted successfully.
    SUBMITTED = 1
    #: Some jobscripts submitted successfully.
    PARTIALLY_SUBMITTED = 2


class Submission(JSONLike):
    """
    A collection of jobscripts to be submitted to a scheduler.

    Parameters
    ----------
    index: int
        The index of this submission.
    jobscripts: list[~hpcflow.app.Jobscript]
        The jobscripts in the submission.
    workflow: ~hpcflow.app.Workflow
        The workflow this is part of.
    submission_parts: dict
        Description of submission parts.
    JS_parallelism: bool
        Whether to exploit jobscript parallelism.
    environments: ~hpcflow.app.EnvironmentsList
        The execution environments to use.
    """

    _child_objects = (
        ChildObjectSpec(
            name="jobscripts",
            class_name="Jobscript",
            is_multiple=True,
            parent_ref="_submission",
        ),
        ChildObjectSpec(
            name="environments",
            class_name="EnvironmentsList",
        ),
    )

    def __init__(
        self,
        index: int,
        jobscripts: List[app.Jobscript],
        workflow: Optional[app.Workflow] = None,
        submission_parts: Optional[Dict] = None,
        JS_parallelism: Optional[bool] = None,
        environments: Optional[app.EnvironmentsList] = None,
    ):
        self._index = index
        self._jobscripts = jobscripts
        self._submission_parts = submission_parts or {}
        self._JS_parallelism = JS_parallelism
        self._environments = environments

        self._submission_parts_lst = None  # assigned on first access; datetime objects

        if workflow:
            #: The workflow this is part of.
            self.workflow = workflow

        self._set_parent_refs()

        for js_idx, js in enumerate(self.jobscripts):
            js._index = js_idx

    @TimeIt.decorator
    def _set_environments(self):
        filterable = ElementResources.get_env_instance_filterable_attributes()

        # map required environments and executable labels to job script indices:
        req_envs = defaultdict(lambda: defaultdict(set))
        for js_idx, js_i in enumerate(self.jobscripts):
            for run in js_i.all_EARs:
                env_spec_h = tuple(zip(*run.env_spec.items()))  # hashable
                for exec_label_j in run.action.get_required_executables():
                    req_envs[env_spec_h][exec_label_j].add(js_idx)
                if env_spec_h not in req_envs:
                    req_envs[env_spec_h] = {}

        # check these envs/execs exist in app data:
        envs = []
        for env_spec_h, exec_js in req_envs.items():
            env_spec = dict(zip(*env_spec_h))
            non_name_spec = {k: v for k, v in env_spec.items() if k != "name"}
            spec_str = f" with specifiers {non_name_spec!r}" if non_name_spec else ""
            env_ref = f"{env_spec['name']!r}{spec_str}"
            try:
                env_i = self.app.envs.get(**env_spec)
            except ObjectListMultipleMatchError:
                raise MultipleEnvironmentsError(
                    f"Multiple environments {env_ref} are defined on this machine."
                )
            except ValueError:
                raise MissingEnvironmentError(
                    f"The environment {env_ref} is not defined on this machine, so the "
                    f"submission cannot be created."
                ) from None
            else:
                if env_i not in envs:
                    envs.append(env_i)

            for exec_i_lab, js_idx_set in exec_js.items():
                try:
                    exec_i = env_i.executables.get(exec_i_lab)
                except ValueError:
                    raise MissingEnvironmentExecutableError(
                        f"The environment {env_ref} as defined on this machine has no "
                        f"executable labelled {exec_i_lab!r}, which is required for this "
                        f"submission, so the submission cannot be created."
                    ) from None

                # check matching executable instances exist:
                for js_idx_j in js_idx_set:
                    js_j = self.jobscripts[js_idx_j]
                    filter_exec = {j: getattr(js_j.resources, j) for j in filterable}
                    exec_instances = exec_i.filter_instances(**filter_exec)
                    if not exec_instances:
                        raise MissingEnvironmentExecutableInstanceError(
                            f"No matching executable instances found for executable "
                            f"{exec_i_lab!r} of environment {env_ref} for jobscript "
                            f"index {js_idx_j!r} with requested resources "
                            f"{filter_exec!r}."
                        )

        # save env definitions to the environments attribute:
        self._environments = self.app.EnvironmentsList(envs)

    def to_dict(self):
        dct = super().to_dict()
        del dct["_workflow"]
        del dct["_index"]
        del dct["_submission_parts_lst"]
        dct = {k.lstrip("_"): v for k, v in dct.items()}
        return dct

    @property
    def index(self) -> int:
        """
        The index of this submission.
        """
        return self._index

    @property
    def environments(self) -> app.EnvironmentsList:
        """
        The execution environments to use.
        """
        return self._environments

    @property
    def submission_parts(self) -> List[Dict]:
        """
        Description of the parts of this submission.
        """
        if not self._submission_parts:
            return []

        if self._submission_parts_lst is None:
            self._submission_parts_lst = [
                {
                    "submit_time": datetime.strptime(dt, self.workflow.ts_fmt)
                    .replace(tzinfo=timezone.utc)
                    .astimezone(),
                    "jobscripts": js_idx,
                }
                for dt, js_idx in self._submission_parts.items()
            ]
        return self._submission_parts_lst

    @TimeIt.decorator
    def get_start_time(self, submit_time: str) -> Union[datetime, None]:
        """Get the start time of a given submission part."""
        js_idx = self._submission_parts[submit_time]
        all_part_starts = []
        for i in js_idx:
            start_time = self.jobscripts[i].start_time
            if start_time:
                all_part_starts.append(start_time)
        if all_part_starts:
            return min(all_part_starts)
        else:
            return None

    @TimeIt.decorator
    def get_end_time(self, submit_time: str) -> Union[datetime, None]:
        """Get the end time of a given submission part."""
        js_idx = self._submission_parts[submit_time]
        all_part_ends = []
        for i in js_idx:
            end_time = self.jobscripts[i].end_time
            if end_time:
                all_part_ends.append(end_time)
        if all_part_ends:
            return max(all_part_ends)
        else:
            return None

    @property
    @TimeIt.decorator
    def start_time(self):
        """Get the first non-None start time over all submission parts."""
        all_start_times = []
        for submit_time in self._submission_parts:
            start_i = self.get_start_time(submit_time)
            if start_i:
                all_start_times.append(start_i)
        if all_start_times:
            return max(all_start_times)
        else:
            return None

    @property
    @TimeIt.decorator
    def end_time(self):
        """Get the final non-None end time over all submission parts."""
        all_end_times = []
        for submit_time in self._submission_parts:
            end_i = self.get_end_time(submit_time)
            if end_i:
                all_end_times.append(end_i)
        if all_end_times:
            return max(all_end_times)
        else:
            return None

    @property
    def jobscripts(self) -> List:
        """
        The jobscripts in this submission.
        """
        return self._jobscripts

    @property
    def JS_parallelism(self):
        """
        Whether to exploit jobscript parallelism.
        """
        return self._JS_parallelism

    @property
    def workflow(self) -> List:
        """
        The workflow this is part of.
        """
        return self._workflow

    @workflow.setter
    def workflow(self, wk):
        self._workflow = wk

    @property
    def jobscript_indices(self) -> Tuple[int]:
        """All associated jobscript indices."""
        return tuple(i.index for i in self.jobscripts)

    @property
    def submitted_jobscripts(self) -> Tuple[int]:
        """Jobscript indices that have been successfully submitted."""
        return tuple(j for i in self.submission_parts for j in i["jobscripts"])

    @property
    def outstanding_jobscripts(self) -> Tuple[int]:
        """Jobscript indices that have not yet been successfully submitted."""
        return tuple(set(self.jobscript_indices) - set(self.submitted_jobscripts))

    @property
    def status(self):
        """
        The status of this submission.
        """
        if not self.submission_parts:
            return SubmissionStatus.PENDING
        else:
            if set(self.submitted_jobscripts) == set(self.jobscript_indices):
                return SubmissionStatus.SUBMITTED
            else:
                return SubmissionStatus.PARTIALLY_SUBMITTED

    @property
    def needs_submit(self):
        """
        Whether this submission needs a submit to be done.
        """
        return self.status in (
            SubmissionStatus.PENDING,
            SubmissionStatus.PARTIALLY_SUBMITTED,
        )

    @property
    def path(self):
        """
        The path to files associated with this submission.
        """
        return self.workflow.submissions_path / str(self.index)

    @property
    def all_EAR_IDs(self):
        """
        The IDs of all EARs in this submission.
        """
        return [i for js in self.jobscripts for i in js.all_EAR_IDs]

    @property
    def all_EARs(self):
        """
        All EARs in this this submission.
        """
        return [i for js in self.jobscripts for i in js.all_EARs]

    @property
    @TimeIt.decorator
    def EARs_by_elements(self):
        """
        All EARs in this submission, grouped by element.
        """
        task_elem_EARs = defaultdict(lambda: defaultdict(list))
        for i in self.all_EARs:
            task_elem_EARs[i.task.index][i.element.index].append(i)
        return task_elem_EARs

    @property
    def abort_EARs_file_name(self):
        """
        The name of a file describing what EARs have aborted.
        """
        return f"abort_EARs.txt"

    @property
    def abort_EARs_file_path(self):
        """
        The path to the file describing what EARs have aborted in this submission.
        """
        return self.path / self.abort_EARs_file_name

    @TimeIt.decorator
    def get_active_jobscripts(
        self, as_json: bool = False
    ) -> List[Tuple[int, Dict[int, JobscriptElementState]]]:
        """Get jobscripts that are active on this machine, and their active states."""
        # this returns: {JS_IDX: {JS_ELEMENT_IDX: STATE}}
        # TODO: query the scheduler once for all jobscripts?
        out = {}
        for js in self.jobscripts:
            active_states = js.get_active_states(as_json=as_json)
            if active_states:
                out[js.index] = active_states
        return out

    def _write_abort_EARs_file(self):
        with self.abort_EARs_file_path.open(mode="wt", newline="\n") as fp:
            # write a single line for each EAR currently in the workflow:
            fp.write("\n".join("0" for _ in range(self.workflow.num_EARs)) + "\n")

    def _set_run_abort(self, run_ID: int):
        """Modify the abort runs file to indicate a specified run should be aborted."""
        with self.abort_EARs_file_path.open(mode="rt", newline="\n") as fp:
            lines = fp.read().splitlines()
        lines[run_ID] = "1"

        # write a new temporary run-abort file:
        tmp_suffix = self.abort_EARs_file_path.suffix + ".tmp"
        tmp = self.abort_EARs_file_path.with_suffix(tmp_suffix)
        self.app.submission_logger.debug(f"Creating temporary run abort file: {tmp!r}.")
        with tmp.open(mode="wt", newline="\n") as fp:
            fp.write("\n".join(i for i in lines) + "\n")

        # atomic rename, overwriting original:
        self.app.submission_logger.debug(
            "Replacing original run abort file with new temporary file."
        )
        os.replace(src=tmp, dst=self.abort_EARs_file_path)

    @staticmethod
    def get_unique_schedulers_of_jobscripts(
        jobscripts: List[Jobscript],
    ) -> Dict[Tuple[Tuple[int, int]], Scheduler]:
        """Get unique schedulers and which of the passed jobscripts they correspond to.

        Uniqueness is determines only by the `Scheduler.unique_properties` tuple.

        Parameters
        ----------
        jobscripts: list[~hpcflow.app.Jobscript]
        """
        js_idx = []
        schedulers = []

        # list of tuples of scheduler properties we consider to determine "uniqueness",
        # with the first string being the scheduler type (class name):
        seen_schedulers = []

        for js in jobscripts:
            if js.scheduler.unique_properties not in seen_schedulers:
                seen_schedulers.append(js.scheduler.unique_properties)
                schedulers.append(js.scheduler)
                js_idx.append([])
            sched_idx = seen_schedulers.index(js.scheduler.unique_properties)
            js_idx[sched_idx].append((js.submission.index, js.index))

        sched_js_idx = dict(zip((tuple(i) for i in js_idx), schedulers))

        return sched_js_idx

    @TimeIt.decorator
    def get_unique_schedulers(self) -> Dict[Tuple[int], Scheduler]:
        """Get unique schedulers and which of this submission's jobscripts they
        correspond to."""
        return self.get_unique_schedulers_of_jobscripts(self.jobscripts)

    @TimeIt.decorator
    def get_unique_shells(self) -> Dict[Tuple[int], Shell]:
        """Get unique shells and which jobscripts they correspond to."""
        js_idx = []
        shells = []

        for js in self.jobscripts:
            if js.shell not in shells:
                shells.append(js.shell)
                js_idx.append([])
            shell_idx = shells.index(js.shell)
            js_idx[shell_idx].append(js.index)

        shell_js_idx = dict(zip((tuple(i) for i in js_idx), shells))

        return shell_js_idx

    def _raise_failure(self, submitted_js_idx, exceptions):
        msg = f"Some jobscripts in submission index {self.index} could not be submitted"
        if submitted_js_idx:
            msg += f" (but jobscripts {submitted_js_idx} were submitted successfully):"
        else:
            msg += ":"

        msg += "\n"
        for sub_err in exceptions:
            msg += (
                f"Jobscript {sub_err.js_idx} at path: {str(sub_err.js_path)!r}\n"
                f"Submit command: {sub_err.submit_cmd!r}.\n"
                f"Reason: {sub_err.message!r}\n"
            )
            if sub_err.subprocess_exc is not None:
                msg += f"Subprocess exception: {sub_err.subprocess_exc}\n"
            if sub_err.job_ID_parse_exc is not None:
                msg += f"Subprocess job ID parse exception: {sub_err.job_ID_parse_exc}\n"
            if sub_err.job_ID_parse_exc is not None:
                msg += f"Job ID parse exception: {sub_err.job_ID_parse_exc}\n"
            if sub_err.stdout:
                msg += f"Submission stdout:\n{indent(sub_err.stdout, '  ')}\n"
            if sub_err.stderr:
                msg += f"Submission stderr:\n{indent(sub_err.stderr, '  ')}\n"

        raise SubmissionFailure(message=msg)

    def _append_submission_part(self, submit_time: str, submitted_js_idx: List[int]):
        self._submission_parts[submit_time] = submitted_js_idx
        self.workflow._store.add_submission_part(
            sub_idx=self.index,
            dt_str=submit_time,
            submitted_js_idx=submitted_js_idx,
        )

    @TimeIt.decorator
    def submit(
        self,
        status,
        ignore_errors: Optional[bool] = False,
        print_stdout: Optional[bool] = False,
        add_to_known: Optional[bool] = True,
    ) -> List[int]:
        """Generate and submit the jobscripts of this submission."""

        # if JS_parallelism explicitly requested but store doesn't support, raise:
        supports_JS_para = self.workflow._store._features.jobscript_parallelism
        if self.JS_parallelism:
            if not supports_JS_para:
                if status:
                    status.stop()
                raise ValueError(
                    f"Store type {self.workflow._store!r} does not support jobscript "
                    f"parallelism."
                )
        elif self.JS_parallelism is None:
            self._JS_parallelism = supports_JS_para

        # set os_name and shell_name for each jobscript:
        for js in self.jobscripts:
            js._set_os_name()
            js._set_shell_name()
            js._set_scheduler_name()

        outstanding = self.outstanding_jobscripts

        # get scheduler, shell and OS version information (also an opportunity to fail
        # before trying to submit jobscripts):
        js_vers_info = {}
        for js_indices, sched in self.get_unique_schedulers().items():
            try:
                vers_info = sched.get_version_info()
            except Exception as err:
                if ignore_errors:
                    vers_info = {}
                else:
                    raise err
            for _, js_idx in js_indices:
                if js_idx in outstanding:
                    if js_idx not in js_vers_info:
                        js_vers_info[js_idx] = {}
                    js_vers_info[js_idx].update(vers_info)

        for js_indices, shell in self.get_unique_shells().items():
            try:
                vers_info = shell.get_version_info()
            except Exception as err:
                if ignore_errors:
                    vers_info = {}
                else:
                    raise err
            for js_idx in js_indices:
                if js_idx in outstanding:
                    if js_idx not in js_vers_info:
                        js_vers_info[js_idx] = {}
                    js_vers_info[js_idx].update(vers_info)

        for js_idx, vers_info_i in js_vers_info.items():
            self.jobscripts[js_idx]._set_version_info(vers_info_i)

        # for direct submission, it's important that os_name/shell_name/scheduler_name
        # are made persistent now, because `Workflow.write_commands`, which might be
        # invoked in a new process before submission has completed, needs to know these:
        self.workflow._store._pending.commit_all()

        # TODO: a submission should only be "submitted" once shouldn't it?
        # no; there could be an IO error (e.g. internet connectivity), so might
        # need to be able to reattempt submission of outstanding jobscripts.
        self.path.mkdir(exist_ok=True)
        if not self.abort_EARs_file_path.is_file():
            self._write_abort_EARs_file()

        # map jobscript `index` to (scheduler job ID or process ID, is_array):
        scheduler_refs = {}
        submitted_js_idx = []
        errs = []
        for js in self.jobscripts:
            # check not previously submitted:
            if js.index not in outstanding:
                continue

            # check all dependencies were submitted now or previously:
            if not all(
                i in submitted_js_idx or i in self.submitted_jobscripts
                for i in js.dependencies
            ):
                continue

            try:
                if status:
                    status.update(f"Submitting jobscript {js.index}...")
                js_ref_i = js.submit(scheduler_refs, print_stdout=print_stdout)
                scheduler_refs[js.index] = (js_ref_i, js.is_array)
                submitted_js_idx.append(js.index)

            except JobscriptSubmissionFailure as err:
                errs.append(err)
                continue

        if submitted_js_idx:
            dt_str = datetime.utcnow().strftime(self.app._submission_ts_fmt)
            self._append_submission_part(
                submit_time=dt_str,
                submitted_js_idx=submitted_js_idx,
            )
            # add a record of the submission part to the known-submissions file
            if add_to_known:
                self.app._add_to_known_submissions(
                    wk_path=self.workflow.path,
                    wk_id=self.workflow.id_,
                    sub_idx=self.index,
                    sub_time=dt_str,
                )

        if errs and not ignore_errors:
            if status:
                status.stop()
            self._raise_failure(submitted_js_idx, errs)

        len_js = len(submitted_js_idx)
        print(f"Submitted {len_js} jobscript{'s' if len_js > 1 else ''}.")

        return submitted_js_idx

    @TimeIt.decorator
    def cancel(self):
        """
        Cancel the active jobs for this submission's jobscripts.
        """
        act_js = list(self.get_active_jobscripts())
        if not act_js:
            print("No active jobscripts to cancel.")
            return
        for js_indices, sched in self.get_unique_schedulers().items():
            # filter by active jobscripts:
            js_idx = [i[1] for i in js_indices if i[1] in act_js]
            if js_idx:
                print(
                    f"Cancelling jobscripts {js_idx!r} of submission {self.index} of "
                    f"workflow {self.workflow.name!r}."
                )
                jobscripts = [self.jobscripts[i] for i in js_idx]
                sched_refs = [i.scheduler_js_ref for i in jobscripts]
                sched.cancel_jobs(js_refs=sched_refs, jobscripts=jobscripts)
            else:
                print("No active jobscripts to cancel.")



from pathlib import Path
import numpy as np
import zarr
import pytest
from hpcflow.sdk.core.test_utils import make_test_data_YAML_workflow
from hpcflow.sdk.persistence.base import StoreEAR, StoreElement, StoreElementIter
from hpcflow.sdk.persistence.json import JSONPersistentStore

from hpcflow.app import app as hf


@pytest.mark.skip("need to refactor `make_test_store_from_spec`")
def test_store_pending_add_task(tmp_path):
    """Check expected pending state after adding a task."""

    # make store: 0 tasks:
    store = JSONPersistentStore.make_test_store_from_spec([], dir=tmp_path)
    task_ID = store.add_task()
    assert store._pending.add_tasks == {task_ID: []}


@pytest.mark.skip("need to refactor `make_test_store_from_spec`")
def test_store_pending_add_element(tmp_path):
    """Check expected pending state after adding an element."""

    # make store: 1 task with 0 elements:
    store = JSONPersistentStore.make_test_store_from_spec(app=hf, spec=[{}], dir=tmp_path)
    elem_ID = store.add_element(task_ID=0)
    assert store._pending.add_elements == {
        elem_ID: StoreElement(
            id_=elem_ID,
            is_pending=True,
            element_idx=0,
            task_ID=0,
            iteration_IDs=[],
        )
    } and store._pending.add_task_element_IDs == {0: [0]}


@pytest.mark.skip("need to refactor `make_test_store_from_spec`")
@pytest.mark.parametrize("elem_ID", [0, 1])
def test_store_pending_add_element_iter(tmp_path, elem_ID):
    """Check expected pending state after adding an element iteration."""

    # make store: 1 task with 2 elements and 0 iterations:
    store = JSONPersistentStore.make_test_store_from_spec(
        [{"elements": [{}, {}]}],
        dir=tmp_path,
    )
    iter_ID = store.add_element_iteration(
        element_ID=elem_ID,
        data_idx={},
        schema_parameters=[],
    )
    assert store._pending.add_elem_iters == {
        iter_ID: StoreElementIter(
            id_=iter_ID,
            is_pending=True,
            element_ID=elem_ID,
            EAR_IDs={},
            data_idx={},
            schema_parameters=[],
        )
    } and store._pending.add_elem_iter_IDs == {elem_ID: [iter_ID]}


@pytest.mark.skip("need to refactor `make_test_store_from_spec`")
def test_store_pending_add_EAR(tmp_path):
    """Check expected pending state after adding an EAR."""

    # make store: 1 task with 1 element and 1 iteration:
    store = JSONPersistentStore.make_test_store_from_spec(
        [{"elements": [{"iterations": [{}]}]}],
        dir=tmp_path,
    )
    EAR_ID = store.add_EAR(
        elem_iter_ID=0,
        action_idx=0,
        commands_idx=[],
        data_idx={},
        metadata={},
    )
    assert store._pending.add_EARs == {
        EAR_ID: StoreEAR(
            id_=EAR_ID,
            is_pending=True,
            elem_iter_ID=0,
            action_idx=0,
            commands_idx=[],
            data_idx={},
            metadata={},
        )
    }


@pytest.mark.skip("need to refactor `make_test_store_from_spec`")
def test_get_task_elements_task_is_pending(tmp_path):
    """Check we get an empty list when getting all task elements of a pending task to
    which no elements have been added."""
    # make store: 0 tasks:
    store = JSONPersistentStore.make_test_store_from_spec([], dir=tmp_path)
    task_ID = store.add_task()
    assert store.get_task_elements(task_ID, slice(0, None)) == []


@pytest.mark.skip("need to refactor `make_test_store_from_spec`")
def test_get_task_elements_single_element_is_pending(tmp_path):
    """Check expected return when getting all task elements of a persistent task that has
    a single pending element."""
    # make store: 1 task
    store = JSONPersistentStore.make_test_store_from_spec([{}], dir=tmp_path)
    store.add_element(task_ID=0)
    assert store.get_task_elements(0, slice(0, None)) == [
        {
            "id": 0,
            "is_pending": True,
            "element_idx": 0,
            "iteration_IDs": [],
            "task_ID": 0,
            "iterations": [],
        }
    ]


@pytest.mark.skip("need to refactor `make_test_store_from_spec`")
def test_get_task_elements_multi_element_one_pending(tmp_path):
    """Check expected return when getting all task elements of a persistent task that has
    a persistent element and a pending element."""
    # make store: 1 task with 1 element:
    store = JSONPersistentStore.make_test_store_from_spec(
        [{"elements": [{}]}], dir=tmp_path
    )
    store.add_element(task_ID=0)
    assert store.get_task_elements(0, slice(0, None)) == [
        {
            "id": 0,
            "is_pending": False,
            "element_idx": 0,
            "iteration_IDs": [],
            "task_ID": 0,
            "iterations": [],
        },
        {
            "id": 1,
            "is_pending": True,
            "element_idx": 1,
            "iteration_IDs": [],
            "task_ID": 0,
            "iterations": [],
        },
    ]


@pytest.mark.skip("need to refactor `make_test_store_from_spec`")
def test_get_task_elements_single_element_iter_pending(tmp_path):
    """Check expected return when getting all task elements of a persistent task that has
    a persistent element with a pending iteration."""
    # make store: 1 task with 1 element:
    store = JSONPersistentStore.make_test_store_from_spec(
        [{"elements": [{}]}], dir=tmp_path
    )
    store.add_element_iteration(element_ID=0, data_idx={}, schema_parameters=[])
    assert store.get_task_elements(0, slice(0, None)) == [
        {
            "id": 0,
            "is_pending": False,
            "element_idx": 0,
            "iteration_IDs": [0],
            "task_ID": 0,
            "iterations": [
                {
                    "id": 0,
                    "is_pending": True,
                    "element_ID": 0,
                    "EAR_IDs": {},
                    "data_idx": {},
                    "schema_parameters": [],
                    "EARs": {},
                }
            ],
        },
    ]


@pytest.mark.skip("need to refactor `make_test_store_from_spec`")
def test_get_task_elements_single_element_iter_EAR_pending(tmp_path):
    """Check expected return when getting all task elements of a persistent task that has
    a persistent element with a persistent iteration and a pending EAR"""
    # make store: 1 task with 1 element with 1 iteration:
    store = JSONPersistentStore.make_test_store_from_spec(
        [{"elements": [{"iterations": [{}]}]}], dir=tmp_path
    )
    store.add_EAR(elem_iter_ID=0, action_idx=0, commands_idx=[], data_idx={}, metadata={})
    assert store.get_task_elements(0, slice(0, None)) == [
        {
            "id": 0,
            "is_pending": False,
            "element_idx": 0,
            "iteration_IDs": [0],
            "task_ID": 0,
            "iterations": [
                {
                    "id": 0,
                    "is_pending": False,
                    "element_ID": 0,
                    "EAR_IDs": {0: [0]},
                    "data_idx": {},
                    "schema_parameters": [],
                    "EARs": {
                        0: [
                            {
                                "id_": 0,
                                "is_pending": True,
                                "elem_iter_ID": 0,
                                "action_idx": 0,
                                "commands_idx": [],
                                "data_idx": {},
                                "metadata": {},
                            }
                        ]
                    },
                },
            ],
        },
    ]


def test_make_zarr_store_zstd_compressor(null_config, tmp_path):
    wk = make_test_data_YAML_workflow(
        workflow_name="workflow_1.yaml",
        path=tmp_path,
        store="zarr",
        store_kwargs={"compressor": "zstd"},
    )


def test_make_zarr_store_no_compressor(null_config, tmp_path):
    wk = make_test_data_YAML_workflow(
        workflow_name="workflow_1.yaml",
        path=tmp_path,
        store="zarr",
        store_kwargs={"compressor": None},
    )


@pytest.mark.integration
def test_zarr_rechunk_data_equivalent(null_config, tmp_path):
    t1 = hf.Task(
        schema=hf.task_schemas.test_t1_conditional_OS,
        inputs={"p1": 101},
        repeats=3,
    )
    wk = hf.Workflow.from_template_data(
        tasks=[t1],
        template_name="test_run_rechunk",
        workflow_name="test_run_rechunk",
        path=tmp_path,
    )
    wk.submit(wait=True, status=False, add_to_known=False)
    wk.rechunk_runs(backup=True, status=False, chunk_size=None)  # None -> one chunk

    arr = wk._store._get_EARs_arr()
    assert arr.chunks == arr.shape

    bak_path = (Path(wk.path) / arr.path).with_suffix(".bak")
    arr_bak = zarr.open(bak_path)

    assert arr_bak.chunks == (1,)

    # check backup and new runs data are equal:
    assert np.all(arr[:] == arr_bak[:])

    # check attributes are equal:
    assert arr.attrs.asdict() == arr_bak.attrs.asdict()


@pytest.mark.integration
def test_zarr_rechunk_data_equivalent_custom_chunk_size(null_config, tmp_path):
    t1 = hf.Task(
        schema=hf.task_schemas.test_t1_conditional_OS,
        inputs={"p1": 101},
        repeats=3,
    )
    wk = hf.Workflow.from_template_data(
        tasks=[t1],
        template_name="test_run_rechunk",
        workflow_name="test_run_rechunk",
        path=tmp_path,
    )
    wk.submit(wait=True, status=False, add_to_known=False)
    wk.rechunk_runs(backup=True, status=False, chunk_size=2)

    arr = wk._store._get_EARs_arr()
    assert arr.chunks == (2,)

    bak_path = (Path(wk.path) / arr.path).with_suffix(".bak")
    arr_bak = zarr.open(bak_path)

    assert arr_bak.chunks == (1,)

    # check backup and new runs data are equal:
    assert np.all(arr[:] == arr_bak[:])


@pytest.mark.integration
def test_zarr_rechunk_data_no_backup_load_runs(null_config, tmp_path):
    t1 = hf.Task(
        schema=hf.task_schemas.test_t1_conditional_OS,
        inputs={"p1": 101},
        repeats=3,
    )
    wk = hf.Workflow.from_template_data(
        tasks=[t1],
        template_name="test_run_rechunk",
        workflow_name="test_run_rechunk",
        path=tmp_path,
    )
    wk.submit(wait=True, status=False, add_to_known=False)
    wk.rechunk_runs(backup=False, status=False)

    arr = wk._store._get_EARs_arr()

    bak_path = (Path(wk.path) / arr.path).with_suffix(".bak")
    assert not bak_path.is_file()

    # check we can load runs:
    runs = wk._store._get_persistent_EARs(id_lst=list(range(wk.num_EARs)))
    run_ID = []
    for i in runs.values():
        run_ID.append(i.id_)


@pytest.mark.integration
def test_zarr_rechunk_data_no_backup_load_parameter_base(null_config, tmp_path):
    t1 = hf.Task(
        schema=hf.task_schemas.test_t1_conditional_OS,
        inputs={"p1": 101},
        repeats=3,
    )
    wk = hf.Workflow.from_template_data(
        tasks=[t1],
        template_name="test_run_rechunk",
        workflow_name="test_run_rechunk",
        path=tmp_path,
    )
    wk.submit(wait=True, status=False, add_to_known=False)
    wk.rechunk_parameter_base(backup=False, status=False)

    arr = wk._store._get_parameter_base_array()

    bak_path = (Path(wk.path) / arr.path).with_suffix(".bak")
    assert not bak_path.is_file()

    # check we can load parameters:
    params = wk.get_all_parameters()
    param_IDs = []
    for i in params:
        param_IDs.append(i.id_)



"""
Dependency resolution cache.
"""

from collections import defaultdict
from dataclasses import dataclass
from typing import Set, Dict

from hpcflow.sdk.log import TimeIt


@dataclass
class DependencyCache:
    """Class to bulk-retrieve dependencies between elements, iterations, and runs."""

    #: What EARs (by ID) a given EAR depends on.
    run_dependencies: Dict[int, Set]
    #: What EARs (by ID) are depending on a given EAR.
    run_dependents: Dict[int, Set]
    #: What EARs (by ID) a given iteration depends on.
    iter_run_dependencies: Dict[int, Set]
    #: What iterations (by ID) a given iteration depends on.
    iter_iter_dependencies: Dict[int, Set]
    #: What iterations (by ID) a given element depends on.
    elem_iter_dependencies: Dict[int, Set]
    #: What elements (by ID) a given element depends on.
    elem_elem_dependencies: Dict[int, Set]
    #: What elements (by ID) are depending on a given element.
    elem_elem_dependents: Dict[int, Set]
    #: Transitive closure of :py:attr:`elem_elem_dependents`.
    elem_elem_dependents_rec: Dict[int, Set]

    #: The elements of the workflow that this cache was built from.
    elements: Dict
    #: The iterations of the workflow that this cache was built from.
    iterations: Dict

    @classmethod
    @TimeIt.decorator
    def build(cls, workflow):
        """
        Build a cache instance.

        Parameters
        ----------
        workflow: ~hpcflow.app.Workflow
            The workflow to build the cache from.
        """
        num_iters = workflow.num_element_iterations
        num_elems = workflow.num_elements
        num_runs = workflow.num_EARs

        all_store_runs = workflow._store.get_EARs(list(range(num_runs)))
        all_store_iters = workflow._store.get_element_iterations(list(range(num_iters)))
        all_store_elements = workflow._store.get_elements(list(range(num_elems)))
        all_param_sources = workflow.get_all_parameter_sources()
        all_data_idx = [
            {
                k: v if isinstance(v, list) else [v]
                for k, v in i.data_idx.items()
                if k not in ("repeats.",)
            }
            for i in all_store_runs
        ]

        # run dependencies and dependents
        run_dependencies = {}
        run_dependents = defaultdict(set)
        for idx, i in enumerate(all_data_idx):
            run_i_sources = set()
            for j in i.values():
                for k in j:
                    run_k = all_param_sources[k].get("EAR_ID")
                    if run_k is not None and run_k != idx:
                        run_i_sources.add(run_k)
            run_dependencies[idx] = run_i_sources
            for m in run_i_sources:
                run_dependents[m].add(idx)

        # add missing:
        for k in range(num_runs):
            run_dependents[k]

        run_dependents = dict(run_dependents)

        # iteration dependencies
        all_iter_run_IDs = {
            i.id_: [k for j in i.EAR_IDs.values() for k in j] for i in all_store_iters
        }
        # for each iteration, which runs does it depend on?
        iter_run_dependencies = {
            k: set(j for i in v for j in run_dependencies[i])
            for k, v in all_iter_run_IDs.items()
        }

        # for each run, which iteration does it belong to?
        all_run_iter_IDs = {}
        for iter_ID, run_IDs in all_iter_run_IDs.items():
            for run_ID in run_IDs:
                all_run_iter_IDs[run_ID] = iter_ID

        # for each iteration, which iterations does it depend on?
        iter_iter_dependencies = {
            k: set(all_run_iter_IDs[i] for i in v)
            for k, v in iter_run_dependencies.items()
        }

        all_elem_iter_IDs = {i.id_: i.iteration_IDs for i in all_store_elements}

        elem_iter_dependencies = {
            k: set(j for i in v for j in iter_iter_dependencies[i])
            for k, v in all_elem_iter_IDs.items()
        }

        # for each iteration, which element does it belong to?
        all_iter_elem_IDs = {}
        for elem_ID, iter_IDs in all_elem_iter_IDs.items():
            for iter_ID in iter_IDs:
                all_iter_elem_IDs[iter_ID] = elem_ID

        # element dependencies
        elem_elem_dependencies = {
            k: set(all_iter_elem_IDs[i] for i in v)
            for k, v in elem_iter_dependencies.items()
        }

        # for each element, which elements depend on it (directly)?
        elem_elem_dependents = defaultdict(set)
        for k, v in elem_elem_dependencies.items():
            for i in v:
                elem_elem_dependents[i].add(k)

        # for each element, which elements depend on it (recursively)?
        elem_elem_dependents_rec = defaultdict(set)
        for k in list(elem_elem_dependents):
            for i in elem_elem_dependents[k]:
                elem_elem_dependents_rec[k].add(i)
                elem_elem_dependents_rec[k].update(
                    {m for m in elem_elem_dependents[i] if m != k}
                )

        # add missing keys:
        for k in range(num_elems):
            elem_elem_dependents[k]
            elem_elem_dependents_rec[k]

        elem_elem_dependents = dict(elem_elem_dependents)
        elem_elem_dependents_rec = dict(elem_elem_dependents_rec)

        elements = workflow.get_all_elements()
        iterations = workflow.get_all_element_iterations()

        return cls(
            run_dependencies=run_dependencies,
            run_dependents=run_dependents,
            iter_run_dependencies=iter_run_dependencies,
            iter_iter_dependencies=iter_iter_dependencies,
            elem_iter_dependencies=elem_iter_dependencies,
            elem_elem_dependencies=elem_elem_dependencies,
            elem_elem_dependents=elem_elem_dependents,
            elem_elem_dependents_rec=elem_elem_dependents_rec,
            elements=elements,
            iterations=iterations,
        )



"""Module defining a function that returns the click CLI group for manipulating the app
configuration."""

import json
import logging
import warnings
from functools import wraps
from contextlib import contextmanager

import click
from colorama import init as colorama_init
from termcolor import colored

from hpcflow.sdk.core import utils

from .errors import ConfigError

logger = logging.getLogger(__name__)

colorama_init(autoreset=True)


def custom_warning_formatter(message, category, filename, lineno, file=None, line=None):
    """Simple warning formatter that shows just the warning type and the message. We use
    this in the CLI, to avoid producing distracting output."""
    return f"{colored(category.__name__, 'yellow')}: {message}\n"


@contextmanager
def warning_formatter(func=custom_warning_formatter):
    """Context manager for modifying the warnings formatter.

    Parameters
    ----------
    func : function to set as the `warnings.formatwarning` function.

    """
    existing_func = warnings.formatwarning
    try:
        warnings.formatwarning = func
        yield
    finally:
        warnings.formatwarning = existing_func


def CLI_exception_wrapper_gen(*exception_cls):
    """
    Decorator factory that enhances the wrapped function to display a nice message on
    success or failure.
    """

    def CLI_exception_wrapper(func):
        """Decorator

        Parameters
        ----------
        func
            Function that return a truthy value if the ???
        """

        @wraps(func)
        @click.pass_context
        def wrapper(ctx, *args, **kwargs):
            try:
                with warning_formatter():
                    out = func(*args, **kwargs)
                if out is not None:
                    click.echo(f"{colored('✔ Config file updated.', 'green')}")
                return out
            except exception_cls as err:
                click.echo(f"{colored(err.__class__.__name__, 'red')}: {err}")
                ctx.exit(1)

        return wrapper

    return CLI_exception_wrapper


def get_config_CLI(app):
    """Generate the configuration CLI for the app."""

    def show_all_config(ctx, param, value):
        if not value or ctx.resilient_parsing:
            return
        ctx.obj["config"]._show(config=True, metadata=False)
        ctx.exit()

    def show_all_metadata(ctx, param, value):
        if not value or ctx.resilient_parsing:
            return
        ctx.obj["config"]._show(config=False, metadata=True)
        ctx.exit()

    def show_config_file(ctx, param, value):
        if not value or ctx.resilient_parsing:
            return
        print(ctx.obj["config"].config_file_contents)
        ctx.exit()

    @click.group()
    @click.option(
        "--no-callback",
        multiple=True,
        help="Exclude a named get/set callback function during execution of the command.",
    )
    @click.pass_context
    def config(ctx, no_callback):
        """Configuration sub-command for getting and setting data in the configuration
        file(s)."""
        ctx.ensure_object(dict)
        ctx.obj["config"] = app.config
        if no_callback:
            ctx.obj["config"]._disable_callbacks(no_callback)

    @config.command("list")
    @click.pass_context
    def config_list(ctx):
        """Show a list of all configurable keys."""
        click.echo("\n".join(ctx.obj["config"].get_configurable()))

    @config.command("import")
    @click.argument("file_path")
    @click.option(
        "--rename/--no-rename",
        default=True,
        help=(
            "Rename the currently loaded config file according to the name of the file "
            "that is being imported (default is to rename). Ignored if `--new` is "
            "specified."
        ),
    )
    @click.option(
        "--new",
        type=click.BOOL,
        is_flag=True,
        default=False,
        help=(
            "If True, generate a new default config, and import the file into this "
            "config. If False, modify the currently loaded config."
        ),
    )
    @click.pass_context
    def import_from_file(ctx, file_path, rename, new):
        """Update the config file with keys from a YAML file."""
        ctx.obj["config"].import_from_file(file_path, rename=rename, make_new=new)

    @config.command()
    @click.argument("name")
    @click.option(
        "--all",
        is_flag=True,
        expose_value=False,
        is_eager=True,
        help="Show all configuration items.",
        callback=show_all_config,
    )
    @click.option(
        "--metadata",
        is_flag=True,
        expose_value=False,
        is_eager=True,
        help="Show all metadata items.",
        callback=show_all_metadata,
    )
    @click.option(
        "--file",
        is_flag=True,
        expose_value=False,
        is_eager=True,
        help="Show the contents of the configuration file.",
        callback=show_config_file,
    )
    @click.pass_context
    @CLI_exception_wrapper_gen(ConfigError)
    def get(ctx, name):
        """Show the value of the specified configuration item."""
        val = ctx.obj["config"].get(name)
        if isinstance(val, list):
            val = "\n".join(str(i) for i in val)
        click.echo(val)

    @config.command()
    @click.argument("name")
    @click.argument("value")
    @click.option(
        "--json",
        "is_json",
        is_flag=True,
        default=False,
        help="Interpret VALUE as a JSON string.",
    )
    @click.pass_context
    @CLI_exception_wrapper_gen(ConfigError)
    def set(ctx, name, value, is_json):
        """Set and save the value of the specified configuration item."""
        ctx.obj["config"].set(name, value, is_json)
        ctx.obj["config"].save()

    @config.command()
    @click.argument("name")
    @click.pass_context
    @CLI_exception_wrapper_gen(ConfigError)
    def unset(ctx, name):
        """Unset and save the value of the specified configuration item."""
        ctx.obj["config"].unset(name)
        ctx.obj["config"].save()

    @config.command()
    @click.argument("name")
    @click.argument("value")
    @click.option(
        "--json",
        "is_json",
        is_flag=True,
        default=False,
        help="Interpret VALUE as a JSON string.",
    )
    @click.pass_context
    @CLI_exception_wrapper_gen(ConfigError)
    def append(ctx, name, value, is_json):
        """Append a new value to the specified configuration item.

        NAME is the dot-delimited path to the list to be appended to.

        """
        ctx.obj["config"].append(name, value, is_json)
        ctx.obj["config"].save()

    @config.command()
    @click.argument("name")
    @click.argument("value")
    @click.option(
        "--json",
        "is_json",
        is_flag=True,
        default=False,
        help="Interpret VALUE as a JSON string.",
    )
    @click.pass_context
    @CLI_exception_wrapper_gen(ConfigError)
    def prepend(ctx, name, value, is_json):
        """Prepend a new value to the specified configuration item.

        NAME is the dot-delimited path to the list to be prepended to.

        """
        ctx.obj["config"].prepend(name, value, is_json)
        ctx.obj["config"].save()

    @config.command(context_settings={"ignore_unknown_options": True})
    @click.argument("name")
    @click.argument("index", type=click.types.INT)
    @click.pass_context
    @CLI_exception_wrapper_gen(ConfigError)
    def pop(ctx, name, index):
        """Remove a value from a list-like configuration item.

        NAME is the dot-delimited path to the list to be modified.

        """
        ctx.obj["config"].pop(name, index)
        ctx.obj["config"].save()

    @config.command()
    @click.argument("name")
    @click.argument("value")
    @click.option(
        "--json",
        "is_json",
        is_flag=True,
        default=False,
        help="Interpret VALUE as a JSON string.",
    )
    @click.pass_context
    @CLI_exception_wrapper_gen(ConfigError)
    def update(ctx, name, value, is_json):
        """Update a map-like value in the configuration.

        NAME is the dot-delimited path to the map to be updated.

        """
        ctx.obj["config"].update(name, value, is_json)
        ctx.obj["config"].save()

    @config.command()
    @click.argument("name")
    @click.option("--defaults")
    @click.pass_context
    @CLI_exception_wrapper_gen(ConfigError)
    def add_scheduler(ctx, name, defaults):
        if defaults:
            defaults = json.loads(defaults)
        else:
            defaults = {}
        ctx.obj["config"].add_scheduler(name, **defaults)
        ctx.obj["config"].save()

    @config.command()
    @click.argument("name")
    @click.option("--defaults")
    @click.pass_context
    @CLI_exception_wrapper_gen(ConfigError)
    def add_shell(ctx, name, defaults):
        if defaults:
            defaults = json.loads(defaults)
        else:
            defaults = {}
        ctx.obj["config"].add_shell(name, **defaults)
        ctx.obj["config"].save()

    @config.command()
    @click.option("--defaults")
    @click.pass_context
    @CLI_exception_wrapper_gen(ConfigError)
    def add_shell_wsl(ctx, defaults):
        if defaults:
            defaults = json.loads(defaults)
        else:
            defaults = {}
        ctx.obj["config"].add_shell_WSL(**defaults)
        ctx.obj["config"].save()

    @config.command()
    @click.argument("sha")
    @click.pass_context
    @CLI_exception_wrapper_gen(ConfigError)
    def set_github_demo_data_dir(ctx, sha):
        ctx.obj["config"].set_github_demo_data_dir(sha=sha)
        ctx.obj["config"].save()

    @config.command()
    def load_data_files():
        """Check we can load the data files (e.g. task schema files) as specified in the
        configuration."""
        app.load_data_files()

    @config.command()
    @click.option("--path", is_flag=True, default=False)
    @click.pass_context
    def open(ctx, path=False):
        """Alias for `{package_name} open config`: open the configuration file, or retrieve
        it's path."""
        file_path = ctx.obj["config"].get("config_file_path")
        if path:
            click.echo(file_path)
        else:
            utils.open_file(file_path)

    @config.command()
    @click.argument("known_name")
    @click.option(
        "--path",
        default=None,
        help=(
            "An `fsspec`-compatible path in which to look for configuration-import "
            "files."
        ),
    )
    @click.pass_context
    def init(ctx, known_name, path):
        ctx.obj["config"].init(known_name=known_name, path=path)

    open.help = open.help.format(package_name=app.package_name)

    return config



"""
Parallel modes.
"""

import enum


class ParallelMode(enum.Enum):
    """
    Potential parallel modes.

    Note: this is not yet implemented in any meaningful way!

    """

    #: Use distributed-memory parallelism (e.g. MPI).
    DISTRIBUTED = 0
    #: Use shared-memory parallelism (e.g. OpenMP).
    SHARED = 1
    #: Use both distributed- and shared-memory parallelism.
    HYBRID = 2



"""Click CLI options that are used as decorators in multiple modules."""

import click

from hpcflow.sdk.core import ALL_TEMPLATE_FORMATS
from hpcflow.sdk.persistence import ALL_STORE_FORMATS, DEFAULT_STORE_FORMAT


def sub_tasks_callback(ctx, param, value):
    """
    Parse subtasks.
    """
    if value:
        return [int(i) for i in value.split(",")]


#: Standard option
format_option = click.option(
    "--format",
    type=click.Choice(ALL_TEMPLATE_FORMATS),
    default=None,
    help=(
        'If specified, one of "json" or "yaml". This forces parsing from a '
        "particular format."
    ),
)
#: Standard option
path_option = click.option(
    "--path",
    type=click.Path(exists=True),
    help="The directory path into which the new workflow will be generated.",
)
#: Standard option
name_option = click.option(
    "--name",
    help=(
        "The name of the workflow. If specified, the workflow directory will be "
        "`path` joined with `name`. If not specified the workflow template name "
        "will be used, in combination with a date-timestamp."
    ),
)
#: Standard option
overwrite_option = click.option(
    "--overwrite",
    is_flag=True,
    default=False,
    help=(
        "If True and the workflow directory (`path` + `name`) already exists, "
        "the existing directory will be overwritten."
    ),
)
#: Standard option
store_option = click.option(
    "--store",
    type=click.Choice(ALL_STORE_FORMATS),
    help="The persistent store type to use.",
    default=DEFAULT_STORE_FORMAT,
)

#: Standard option
ts_fmt_option = click.option(
    "--ts-fmt",
    help=(
        "The datetime format to use for storing datetimes. Datetimes are always "
        "stored in UTC (because Numpy does not store time zone info), so this "
        "should not include a time zone name."
    ),
)
#: Standard option
ts_name_fmt_option = click.option(
    "--ts-name-fmt",
    help=(
        "The datetime format to use when generating the workflow name, where it "
        "includes a timestamp."
    ),
)

#: Standard option
variables_option = click.option(
    "-v",
    "--var",
    "variables",
    type=(str, str),
    multiple=True,
    help=(
        "Workflow template variable value to be substituted in to the template file or "
        "string. Multiple variable values can be specified."
    ),
)
#: Standard option
js_parallelism_option = click.option(
    "--js-parallelism",
    help=(
        "If True, allow multiple jobscripts to execute simultaneously. Raises if "
        "set to True but the store type does not support the "
        "`jobscript_parallelism` feature. If not set, jobscript parallelism will "
        "be used if the store type supports it."
    ),
    type=click.BOOL,
)
#: Standard option
wait_option = click.option(
    "--wait",
    help=("If True, this command will block until the workflow execution is complete."),
    is_flag=True,
    default=False,
)
#: Standard option
add_to_known_opt = click.option(
    "--add-to-known/--no-add-to-known",
    default=True,
    help="If True, add this submission to the known-submissions file.",
)
#: Standard option
print_idx_opt = click.option(
    "--print-idx",
    help="If True, print the submitted jobscript indices for each submission index.",
    is_flag=True,
    default=False,
)
#: Standard option
tasks_opt = click.option(
    "--tasks",
    help=(
        "List of comma-separated task indices to include in this submission. By default "
        "all tasks are included."
    ),
    callback=sub_tasks_callback,
)
#: Standard option
cancel_opt = click.option(
    "--cancel",
    help="Immediately cancel the submission. Useful for testing and benchmarking.",
    is_flag=True,
    default=False,
)
#: Standard option
submit_status_opt = click.option(
    "--status/--no-status",
    help="If True, display a live status to track submission progress.",
    default=True,
)
#: Standard option
make_status_opt = click.option(
    "--status/--no-status",
    help="If True, display a live status to track workflow creation progress.",
    default=True,
)

#: Standard option
zip_path_opt = click.option(
    "--path",
    default=".",
    help=(
        "Path at which to create the new zipped workflow. If this is an existing "
        "directory, the zip file will be created within this directory. Otherwise, this "
        "path is assumed to be the full file path to the new zip file."
    ),
)
#: Standard option
zip_overwrite_opt = click.option(
    "--overwrite",
    is_flag=True,
    default=False,
    help="If set, any existing file will be overwritten.",
)
#: Standard option
zip_log_opt = click.option("--log", help="Path to a log file to use during zipping.")
#: Standard option
zip_include_execute_opt = click.option("--include-execute", is_flag=True)
#: Standard option
zip_include_rechunk_backups_opt = click.option("--include-rechunk-backups", is_flag=True)

#: Standard option
unzip_path_opt = click.option(
    "--path",
    default=".",
    help=(
        "Path at which to create the new unzipped workflow. If this is an existing "
        "directory, the new workflow directory will be created within this directory. "
        "Otherwise, this path will represent the new workflow directory path."
    ),
)
#: Standard option
unzip_log_opt = click.option("--log", help="Path to a log file to use during unzipping.")

#: Standard option
rechunk_backup_opt = click.option(
    "--backup/--no-backup",
    default=True,
    help=("First copy a backup of the array to a directory ending in `.bak`."),
)
#: Standard option
rechunk_chunk_size_opt = click.option(
    "--chunk-size",
    type=click.INT,
    default=None,
    help=(
        "New chunk size (array items per chunk). If unset (as by default), the array "
        "will be rechunked to a single chunk array (i.e with a chunk size equal to the "
        "array's shape)."
    ),
)
#: Standard option
rechunk_status_opt = click.option(
    "--status/--no-status",
    default=True,
    help="If True, display a live status to track rechunking progress.",
)


def _add_doc_from_help(*args):
    """
    Attach the ``help`` field of each of its arguments as its ``__doc__``.
    Only necessary because the wrappers in Click don't do this for us.

    :meta private:
    """
    # Yes, this is ugly!
    from types import SimpleNamespace

    for opt in args:
        ns = SimpleNamespace()
        params = getattr(opt(ns), "__click_params__", [])
        if params:
            help = getattr(params[0], "help", "")
            if help:
                opt.__doc__ = f"Click option decorator: {help}"


_add_doc_from_help(
    format_option,
    path_option,
    name_option,
    overwrite_option,
    store_option,
    ts_fmt_option,
    ts_name_fmt_option,
    variables_option,
    js_parallelism_option,
    wait_option,
    add_to_known_opt,
    print_idx_opt,
    tasks_opt,
    cancel_opt,
    submit_status_opt,
    make_status_opt,
    zip_path_opt,
    zip_overwrite_opt,
    zip_log_opt,
    zip_include_execute_opt,
    zip_include_rechunk_backups_opt,
    unzip_path_opt,
    unzip_log_opt,
    rechunk_backup_opt,
    rechunk_chunk_size_opt,
    rechunk_status_opt,
)



"""
Adapters for various shells.
"""
import os
from typing import Dict, Optional

from hpcflow.sdk.core.errors import UnsupportedShellError

from .base import Shell
from .bash import Bash, WSLBash
from .powershell import WindowsPowerShell

#: All supported shells.
ALL_SHELLS = {
    "bash": {"posix": Bash},
    "powershell": {"nt": WindowsPowerShell},
    "wsl+bash": {"nt": WSLBash},
    "wsl": {"nt": WSLBash},  # TODO: cast this to wsl+bash in ResourceSpec?
}

#: The default shell in the default config.
DEFAULT_SHELL_NAMES = {
    "posix": "bash",
    "nt": "powershell",
}


def get_supported_shells(os_name: Optional[str] = None) -> Dict[str, Shell]:
    """
    Get shells supported on the current or given OS.
    """
    os_name = os_name or os.name
    return {k: v.get(os_name) for k, v in ALL_SHELLS.items() if v.get(os_name)}


def get_shell(shell_name, os_name: Optional[str] = None, **kwargs) -> Shell:
    """
    Get a shell interface with the given name for a given OS (or the current one).
    """
    # TODO: apply config default shell args?

    os_name = os_name or os.name
    shell_name = shell_name.lower()

    supported = get_supported_shells(os_name.lower())
    shell_cls = supported.get(shell_name)
    if not shell_cls:
        raise UnsupportedShellError(shell=shell_name, supported=supported)

    shell_obj = shell_cls(**kwargs)

    return shell_obj



"""
Model of files in the run directory.
"""

import re
from hpcflow.sdk.core.utils import JSONLikeDirSnapShot


class RunDirAppFiles:
    """A class to encapsulate the naming/recognition of app-created files within run
    directories."""

    _app_attr = "app"

    _CMD_FILES_RE_PATTERN = r"js_\d+_act_\d+\.?\w*"

    @classmethod
    def get_log_file_name(cls):
        """File name for the app log file."""
        return f"{cls.app.package_name}.log"

    @classmethod
    def get_std_file_name(cls):
        """File name for stdout and stderr streams from the app."""
        return f"{cls.app.package_name}_std.txt"

    @staticmethod
    def get_run_file_prefix(js_idx: int, js_action_idx: int):
        """
        Get the common prefix for files associated with a run.
        """
        return f"js_{js_idx}_act_{js_action_idx}"

    @classmethod
    def get_commands_file_name(cls, js_idx: int, js_action_idx: int, shell):
        """
        Get the name of the file containing commands.
        """
        return cls.get_run_file_prefix(js_idx, js_action_idx) + shell.JS_EXT

    @classmethod
    def get_run_param_dump_file_prefix(cls, js_idx: int, js_action_idx: int):
        """Get the prefix to a file in the run directory that the app will dump parameter
        data to."""
        return cls.get_run_file_prefix(js_idx, js_action_idx) + "_inputs"

    @classmethod
    def get_run_param_load_file_prefix(cls, js_idx: int, js_action_idx: int):
        """Get the prefix to a file in the run directory that the app will load parameter
        data from."""
        return cls.get_run_file_prefix(js_idx, js_action_idx) + "_outputs"

    @classmethod
    def take_snapshot(cls):
        """Take a JSONLikeDirSnapShot, and process to ignore files created by the app.

        This includes command files that are invoked by jobscripts, the app log file, and
        the app standard out/error file.

        """
        snapshot = JSONLikeDirSnapShot()
        snapshot.take(".")
        ss_js = snapshot.to_json_like()
        ss_js.pop("root_path")  # always the current working directory of the run
        for k in list(ss_js["data"].keys()):
            if (
                k == cls.get_log_file_name()
                or k == cls.get_std_file_name()
                or re.match(cls._CMD_FILES_RE_PATTERN, k)
            ):
                ss_js["data"].pop(k)

        return ss_js



import pytest
from hpcflow.app import app as hf
from hpcflow.sdk.core.errors import MissingElementGroup


def test_group_simple(null_config, tmp_path):
    s1 = hf.TaskSchema(
        objective="t1",
        inputs=[hf.SchemaInput("p1")],
        outputs=[hf.SchemaOutput("p2")],
        actions=[
            hf.Action(
                commands=[
                    hf.Command(
                        "echo $(( <<parameter:p1>> + 1 ))", stdout="<<parameter:p2>>"
                    )
                ]
            )
        ],
    )
    s2 = hf.TaskSchema(
        objective="t2",
        inputs=[hf.SchemaInput("p2", group="my_group")],
        outputs=[hf.SchemaOutput("p3")],
        actions=[
            hf.Action(
                commands=[
                    hf.Command(
                        "echo $(( <<parameter:p2>> + 2 ))", stdout="<<parameter:p3>>"
                    )
                ]
            )
        ],
    )

    t1 = hf.Task(
        schema=s1,
        sequences=[hf.ValueSequence("inputs.p1", values=[1, 2, 3])],
        groups=[hf.ElementGroup(name="my_group")],
    )
    t2 = hf.Task(schema=s2)
    wk = hf.Workflow.from_template_data(
        template_name="test_groups",
        path=tmp_path,
        tasks=[t1, t2],
    )
    assert [task.num_elements for task in wk.tasks] == [3, 1]
    assert len(wk.tasks.t2.elements[0].get_data_idx("inputs.p2")["inputs.p2"]) == 3


def test_group_raise_no_elements(null_config, tmp_path):
    s1 = hf.TaskSchema(
        objective="t1",
        inputs=[hf.SchemaInput("p1")],
        outputs=[hf.SchemaOutput("p2")],
        actions=[
            hf.Action(
                commands=[
                    hf.Command(
                        "echo $(( <<parameter:p1>> + 1 ))", stdout="<<parameter:p2>>"
                    )
                ]
            )
        ],
    )
    s2 = hf.TaskSchema(
        objective="t2",
        inputs=[hf.SchemaInput("p2", group="my_group")],
        outputs=[hf.SchemaOutput("p3")],
        actions=[
            hf.Action(
                commands=[
                    hf.Command(
                        "echo $(( <<parameter:p2>> + 2 ))", stdout="<<parameter:p3>>"
                    )
                ]
            )
        ],
    )

    t1 = hf.Task(
        schema=s1,
        sequences=[hf.ValueSequence("inputs.p1", values=[1, 2, 3])],
    )
    t2 = hf.Task(schema=s2)
    with pytest.raises(MissingElementGroup):
        hf.Workflow.from_template_data(
            template_name="test_groups",
            path=tmp_path,
            tasks=[t1, t2],
        )



"""
Job scheduler models.
"""

from pathlib import Path
import sys
import time
from typing import Any, List, Tuple


class NullScheduler:
    """
    Abstract base class for schedulers.

    Keyword Args
    ------------
    shell_args: str
        Arguments to pass to the shell. Pre-quoted.
    shebang_args: str
        Arguments to set on the shebang line. Pre-quoted.
    options: dict
        Options to the scheduler.
    """

    #: Default value for arguments to the shell.
    DEFAULT_SHELL_ARGS = ""
    #: Default value for arguments on the shebang line.
    DEFAULT_SHEBANG_ARGS = ""

    def __init__(
        self,
        shell_args=None,
        shebang_args=None,
        options=None,
    ):
        self.shebang_args = shebang_args or self.DEFAULT_SHEBANG_ARGS
        self.shell_args = shell_args or self.DEFAULT_SHELL_ARGS
        self.options = options or {}

    @property
    def unique_properties(self):
        """
        Unique properties, for hashing.
        """
        return (self.__class__.__name__,)

    def __eq__(self, other) -> bool:
        if type(self) != type(other):
            return False
        else:
            return self.__dict__ == other.__dict__

    def get_version_info(self):
        """
        Get the version of the scheduler.
        """
        return {}

    def parse_submission_output(self, stdout: str) -> None:
        """
        Parse the output from a submission to determine the submission ID.
        """
        return None

    @staticmethod
    def is_num_cores_supported(num_cores, core_range: List[int]):
        """
        Test whether particular number of cores is supported in given range of cores.
        """
        step = core_range[1] if core_range[1] is not None else 1
        upper = core_range[2] + 1 if core_range[2] is not None else sys.maxsize
        return num_cores in range(core_range[0], upper, step)


class Scheduler(NullScheduler):
    """
    Base class for schedulers that use a job submission system.

    Parameters
    ----------
    submit_cmd: str
        The submission command, if overridden from default.
    show_cmd: str
        The show command, if overridden from default.
    del_cmd: str
        The delete command, if overridden from default.
    js_cmd: str
        The job script command, if overridden from default.
    login_nodes_cmd: str
        The login nodes command, if overridden from default.
    array_switch: str
        The switch to enable array jobs, if overridden from default.
    array_item_var: str
        The variable for array items, if overridden from default.
    """

    #: Default command for logging into nodes.
    DEFAULT_LOGIN_NODES_CMD = None
    #: Default pattern for matching the names of login nodes.
    DEFAULT_LOGIN_NODE_MATCH = "*login*"

    def __init__(
        self,
        submit_cmd=None,
        show_cmd=None,
        del_cmd=None,
        js_cmd=None,
        login_nodes_cmd=None,
        array_switch=None,
        array_item_var=None,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        self.submit_cmd = submit_cmd or self.DEFAULT_SUBMIT_CMD
        self.show_cmd = show_cmd or self.DEFAULT_SHOW_CMD
        self.del_cmd = del_cmd or self.DEFAULT_DEL_CMD
        self.js_cmd = js_cmd or self.DEFAULT_JS_CMD
        self.login_nodes_cmd = login_nodes_cmd or self.DEFAULT_LOGIN_NODES_CMD
        self.array_switch = array_switch or self.DEFAULT_ARRAY_SWITCH
        self.array_item_var = array_item_var or self.DEFAULT_ARRAY_ITEM_VAR

    @property
    def unique_properties(self):
        return (self.__class__.__name__, self.submit_cmd, self.show_cmd, self.del_cmd)

    def format_switch(self, switch):
        """
        Format a particular switch to use the JS command.
        """
        return f"{self.js_cmd} {switch}"

    def is_jobscript_active(self, job_ID: str):
        """Query if a jobscript is running/pending."""
        return bool(self.get_job_state_info([job_ID]))

    def wait_for_jobscripts(self, js_refs: List[Any]) -> None:
        """
        Wait for jobscripts to update their state.
        """
        while js_refs:
            info = self.get_job_state_info(js_refs)
            print(info)
            if not info:
                break
            js_refs = list(info.keys())
            time.sleep(2)



"""
Interface to the standard logger, and performance logging utility.
"""

from functools import wraps
import logging
from pathlib import Path
import time
from collections import defaultdict
import statistics


class TimeIt:
    """
    Method execution time instrumentation.
    """

    #: Whether the instrumentation is active.
    active = False
    #: Where to log to.
    file_path = None
    #: The details be tracked.
    timers = defaultdict(list)
    #: Traces of the stack.
    trace = []
    #: Trace indices.
    trace_idx = []
    #: Preceding traces.
    trace_prev = []
    #: Preceding trace indices.
    trace_idx_prev = []

    @classmethod
    def decorator(cls, func):
        """
        Decorator for a method that is to have its execution time monitored.
        """

        @wraps(func)
        def wrapper(*args, **kwargs):

            if not cls.active:
                return func(*args, **kwargs)

            cls.trace.append(func.__qualname__)

            if cls.trace_prev == cls.trace:
                new_trace_idx = cls.trace_idx_prev[-1] + 1
            else:
                new_trace_idx = 0
            cls.trace_idx.append(new_trace_idx)

            tic = time.perf_counter()
            out = func(*args, **kwargs)
            toc = time.perf_counter()
            elapsed = toc - tic

            cls.timers[tuple(cls.trace)].append(elapsed)

            cls.trace_prev = list(cls.trace)
            cls.trace_idx_prev = list(cls.trace_idx)

            cls.trace.pop()
            cls.trace_idx.pop()

            return out

        return wrapper

    @classmethod
    def summarise(cls):
        """
        Produce a machine-readable summary of method execution time statistics.
        """
        stats = {
            k: {
                "number": len(v),
                "mean": statistics.mean(v),
                "stddev": statistics.pstdev(v),
                "min": min(v),
                "max": max(v),
                "sum": sum(v),
            }
            for k, v in cls.timers.items()
        }

        # make a graph
        for k in stats:
            stats[k]["children"] = {}

        for key in sorted(stats.keys(), key=lambda x: len(x), reverse=True):
            if len(key) == 1:
                continue
            value = stats.pop(key)
            parent = key[:-1]
            for other_key in stats.keys():
                if other_key == parent:
                    stats[other_key]["children"][key] = value
                    break

        return stats

    @classmethod
    def summarise_string(cls):
        """
        Produce a human-readable summary of method execution time statistics.
        """

        def _format_nodes(node, depth=0, depth_final=None):
            if depth_final is None:
                depth_final = []
            out = []
            for idx, (k, v) in enumerate(node.items()):
                is_final_child = idx == len(node) - 1
                angle = "└ " if is_final_child else "├ "
                bars = ""
                if depth > 0:
                    bars = "".join(f"{'│ ' if not i else '  '}" for i in depth_final)
                k_str = bars + (angle if depth > 0 else "") + f"{k[depth]}"
                number = v["number"]
                min_str = f"{v['min']:10.6f}" if number > 1 else f"{f'-':^12s}"
                max_str = f"{v['max']:10.6f}" if number > 1 else f"{f'-':^12s}"
                stddev_str = f"({v['stddev']:8.6f})" if number > 1 else f"{f' ':^10s}"
                out.append(
                    f"{k_str:.<80s} {v['sum']:12.6f} "
                    f"{v['mean']:10.6f} {stddev_str} {number:8d} "
                    f"{min_str} {max_str} "
                )
                depth_final_next = list(depth_final) + (
                    [is_final_child] if depth > 0 else []
                )
                out.extend(
                    _format_nodes(
                        v["children"], depth=depth + 1, depth_final=depth_final_next
                    )
                )
            return out

        summary = cls.summarise()

        out = [
            f"{'function':^80s} {'sum /s':^12s} {'mean (stddev) /s':^20s} {'N':^8s} "
            f"{'min /s':^12s} {'max /s':^12s}"
        ]
        out += _format_nodes(summary)
        out_str = "\n".join(out)
        if cls.file_path:
            Path(cls.file_path).write_text(out_str, encoding="utf-8")
        else:
            print(out_str)


class AppLog:
    """
    Application log control.
    """

    #: Default logging level for the console.
    DEFAULT_LOG_CONSOLE_LEVEL = "WARNING"
    #: Default logging level for log files.
    DEFAULT_LOG_FILE_LEVEL = "INFO"

    def __init__(self, app, log_console_level=None):
        #: The application context.
        self.app = app
        #: The base logger for the application.
        self.logger = logging.getLogger(app.package_name)
        self.logger.setLevel(logging.DEBUG)
        #: The handler for directing logging messages to the console.
        self.console_handler = self._add_console_logger(
            level=log_console_level or AppLog.DEFAULT_LOG_CONSOLE_LEVEL
        )

    def _add_console_logger(self, level, fmt=None):
        fmt = fmt or "%(levelname)s %(name)s: %(message)s"
        handler = logging.StreamHandler()
        handler.setFormatter(logging.Formatter(fmt))
        handler.setLevel(level)
        self.logger.addHandler(handler)
        return handler

    def update_console_level(self, new_level):
        """
        Set the logging level for console messages.
        """
        if new_level:
            self.console_handler.setLevel(new_level.upper())

    def add_file_logger(self, path, level=None, fmt=None, max_bytes=None):
        """
        Add a log file.
        """
        fmt = fmt or f"%(asctime)s %(levelname)s %(name)s: %(message)s"
        level = level or AppLog.DEFAULT_LOG_FILE_LEVEL
        max_bytes = max_bytes or int(10e6)

        if not path.parent.is_dir():
            self.logger.info(f"Generating log file parent directory: {path.parent!r}")
            path.parent.mkdir(exist_ok=True, parents=True)

        handler = logging.handlers.RotatingFileHandler(filename=path, maxBytes=max_bytes)
        handler.setFormatter(logging.Formatter(fmt))
        handler.setLevel(level.upper())
        self.logger.addHandler(handler)
        return handler

    def remove_file_handlers(self):
        """Remove all file handlers."""
        # TODO: store a `file_handlers` attribute as well as `console_handlers`
        for hdlr in self.logger.handlers:
            if isinstance(hdlr, logging.FileHandler):
                self.logger.debug(f"Removing file handler from the AppLog: {hdlr!r}.")
                self.logger.removeHandler(hdlr)



"""
Errors from the workflow system.
"""

import os
from typing import Iterable, List


class InputValueDuplicateSequenceAddress(ValueError):
    """
    An InputValue has the same sequence address twice.
    """


class TaskTemplateMultipleSchemaObjectives(ValueError):
    """
    A TaskTemplate has multiple objectives.
    """


class TaskTemplateUnexpectedInput(ValueError):
    """
    A TaskTemplate was given unexpected input.
    """


class TaskTemplateUnexpectedSequenceInput(ValueError):
    """
    A TaskTemplate was given an unexpected sequence.
    """


class TaskTemplateMultipleInputValues(ValueError):
    """
    A TaskTemplate had multiple input values bound over each other.
    """


class InvalidIdentifier(ValueError):
    """
    A bad identifier name was given.
    """


class MissingInputs(Exception):
    """
    Inputs were missing.

    Parameters
    ----------
    message: str
        The message of the exception.
    missing_inputs: list[str]
        The missing inputs.
    """

    # TODO: add links to doc pages for common user-exceptions?

    def __init__(self, message, missing_inputs) -> None:
        self.missing_inputs = missing_inputs
        super().__init__(message)


class UnrequiredInputSources(ValueError):
    """
    Input sources were provided that were not required.

    Parameters
    ----------
    message: str
        The message of the exception.
    unrequired_sources: list
        The input sources that were not required.
    """

    def __init__(self, message, unrequired_sources) -> None:
        self.unrequired_sources = unrequired_sources
        for src in unrequired_sources:
            if src.startswith("inputs."):
                # reminder about how to specify input sources:
                message += (
                    f" Note that input source keys should not be specified with the "
                    f"'inputs.' prefix. Did you mean to specify {src[len('inputs.'):]!r} "
                    f"instead of {src!r}?"
                )
                break
        super().__init__(message)


class ExtraInputs(Exception):
    """
    Extra inputs were provided.

    Parameters
    ----------
    message: str
        The message of the exception.
    extra_inputs: list
        The extra inputs.
    """

    def __init__(self, message, extra_inputs) -> None:
        self.extra_inputs = extra_inputs
        super().__init__(message)


class UnavailableInputSource(ValueError):
    """
    An input source was not available.
    """


class InapplicableInputSourceElementIters(ValueError):
    """
    An input source element iteration was inapplicable."""


class NoCoincidentInputSources(ValueError):
    """
    Could not line up input sources to make an actual valid execution.
    """


class TaskTemplateInvalidNesting(ValueError):
    """
    Invalid nesting in a task template.
    """


class TaskSchemaSpecValidationError(Exception):
    """
    A task schema failed to validate.
    """


class WorkflowSpecValidationError(Exception):
    """
    A workflow failed to validate.
    """


class InputSourceValidationError(Exception):
    """
    An input source failed to validate.
    """


class EnvironmentSpecValidationError(Exception):
    """
    An environment specification failed to validate.
    """


class ParameterSpecValidationError(Exception):
    """
    A parameter specification failed to validate.
    """


class FileSpecValidationError(Exception):
    """
    A file specification failed to validate.
    """


class DuplicateExecutableError(ValueError):
    """
    The same executable was present twice in an executable environment.
    """


class MissingCompatibleActionEnvironment(Exception):
    """
    Could not find a compatible action environment.
    """


class MissingActionEnvironment(Exception):
    """
    Could not find an action environment.
    """


class ActionEnvironmentMissingNameError(Exception):
    """
    An action environment was missing its name.
    """


class FromSpecMissingObjectError(Exception):
    """
    Missing object when deserialising from specification.
    """


class TaskSchemaMissingParameterError(Exception):
    """
    Parameter was missing from task schema.
    """


class ToJSONLikeChildReferenceError(Exception):
    """
    Failed to generate or reference a child object when converting to JSON.
    """


class InvalidInputSourceTaskReference(Exception):
    """
    Invalid input source in task reference.
    """


class WorkflowNotFoundError(Exception):
    """
    Could not find the workflow.
    """


class MalformedWorkflowError(Exception):
    """
    Workflow was a malformed document.
    """


class ValuesAlreadyPersistentError(Exception):
    """
    Trying to make a value persistent that already is so.
    """


class MalformedParameterPathError(ValueError):
    """
    The path to a parameter was ill-formed.
    """


class MalformedNestingOrderPath(ValueError):
    """
    A nesting order path was ill-formed.
    """


class UnknownResourceSpecItemError(ValueError):
    """
    A resource specification item was not found.
    """


class WorkflowParameterMissingError(AttributeError):
    """
    A parameter to a workflow was missing.
    """


class WorkflowBatchUpdateFailedError(Exception):
    """
    An update to a workflow failed.
    """


class WorkflowLimitsError(ValueError):
    """
    Workflow hit limits.
    """


class UnsetParameterDataError(Exception):
    """
    Tried to read from an unset parameter.
    """


class LoopAlreadyExistsError(Exception):
    """
    A particular loop (or its name) already exists.
    """


class LoopTaskSubsetError(ValueError):
    """
    Problem constructing a subset of a task for a loop.
    """


class SchedulerVersionsFailure(RuntimeError):
    """We couldn't get the scheduler and or shell versions."""

    def __init__(self, message):
        self.message = message
        super().__init__(message)


class JobscriptSubmissionFailure(RuntimeError):
    """
    A job script could not be submitted to the scheduler.
    """

    def __init__(
        self,
        message,
        submit_cmd,
        js_idx,
        js_path,
        stdout,
        stderr,
        subprocess_exc,
        job_ID_parse_exc,
    ) -> None:
        self.message = message
        self.submit_cmd = submit_cmd
        self.js_idx = js_idx
        self.js_path = js_path
        self.stdout = stdout
        self.stderr = stderr
        self.subprocess_exc = subprocess_exc
        self.job_ID_parse_exc = job_ID_parse_exc
        super().__init__(message)


class SubmissionFailure(RuntimeError):
    """
    A job submission failed.
    """

    def __init__(self, message) -> None:
        self.message = message
        super().__init__(message)


class WorkflowSubmissionFailure(RuntimeError):
    """
    A workflow submission failed.
    """


class ResourceValidationError(ValueError):
    """An incompatible resource requested by the user."""


class UnsupportedOSError(ResourceValidationError):
    """This machine is not of the requested OS."""

    def __init__(self, os_name) -> None:
        message = (
            f"OS {os_name!r} is not compatible with this machine/instance with OS: "
            f"{os.name!r}."
        )
        super().__init__(message)
        self.os_name = os_name


class UnsupportedShellError(ResourceValidationError):
    """We don't support this shell on this OS."""

    def __init__(self, shell, supported) -> None:
        message = (
            f"Shell {shell!r} is not supported on this machine/instance. Supported "
            f"shells are: {supported!r}."
        )
        super().__init__(message)
        self.shell = shell
        self.supported = supported


class UnsupportedSchedulerError(ResourceValidationError):
    """This scheduler is not supported on this machine according to the config.

    This is also raised in config validation when attempting to add a scheduler that is
    not known for this OS.

    """

    def __init__(self, scheduler, supported=None, available=None) -> None:
        if supported is not None:
            message = (
                f"Scheduler {scheduler!r} is not supported on this machine/instance. "
                f"Supported schedulers according to the app configuration are: "
                f"{supported!r}."
            )
        elif available is not None:
            message = (
                f"Scheduler {scheduler!r} is not supported on this OS. Schedulers "
                f"compatible with this OS are: {available!r}."
            )
        super().__init__(message)
        self.scheduler = scheduler
        self.supported = supported
        self.available = available


class UnknownSGEPEError(ResourceValidationError):
    """
    Miscellaneous error from SGE parallel environment.
    """


class IncompatibleSGEPEError(ResourceValidationError):
    """
    The SGE parallel environment selected is incompatible.
    """


class NoCompatibleSGEPEError(ResourceValidationError):
    """
    No SGE parallel environment is compatible with request.
    """


class IncompatibleParallelModeError(ResourceValidationError):
    """
    The parallel mode is incompatible.
    """


class UnknownSLURMPartitionError(ResourceValidationError):
    """
    The requested SLURM partition isn't known.
    """


class IncompatibleSLURMPartitionError(ResourceValidationError):
    """
    The requested SLURM partition is incompatible.
    """


class IncompatibleSLURMArgumentsError(ResourceValidationError):
    """
    The SLURM arguments are incompatible with each other.
    """


class _MissingStoreItemError(ValueError):
    def __init__(self, id_lst: Iterable[int], item_type: str) -> None:
        message = (
            f"Store {item_type}s with the following IDs do not all exist: {id_lst!r}"
        )
        super().__init__(message)
        self.id_lst = id_lst


class MissingStoreTaskError(_MissingStoreItemError):
    """Some task IDs do not exist."""

    _item_type = "task"

    def __init__(self, id_lst: Iterable[int]) -> None:
        super().__init__(id_lst, self._item_type)


class MissingStoreElementError(_MissingStoreItemError):
    """Some element IDs do not exist."""

    _item_type = "element"

    def __init__(self, id_lst: Iterable[int]) -> None:
        super().__init__(id_lst, self._item_type)


class MissingStoreElementIterationError(_MissingStoreItemError):
    """Some element iteration IDs do not exist."""

    _item_type = "element iteration"

    def __init__(self, id_lst: Iterable[int]) -> None:
        super().__init__(id_lst, self._item_type)


class MissingStoreEARError(_MissingStoreItemError):
    """Some EAR IDs do not exist."""

    _item_type = "EAR"

    def __init__(self, id_lst: Iterable[int]) -> None:
        super().__init__(id_lst, self._item_type)


class MissingParameterData(_MissingStoreItemError):
    """Some parameter IDs do not exist"""

    _item_type = "parameter"

    def __init__(self, id_lst: Iterable[int]) -> None:
        super().__init__(id_lst, self._item_type)


class NotSubmitMachineError(RuntimeError):
    """
    The requested machine can't be submitted to.
    """


class RunNotAbortableError(ValueError):
    """
    Cannot abort the run.
    """


class NoCLIFormatMethodError(AttributeError):
    """
    Some CLI class lacks a format method
    """


class ContainerKeyError(KeyError):
    """
    A key could not be mapped in a container.

    Parameters
    ----------
    path: list[str]
        The path whose resolution failed.
    """

    def __init__(self, path: List[str]) -> None:
        self.path = path
        super().__init__()


class MayNeedObjectError(Exception):
    """
    An object is needed but not present.

    Parameters
    ----------
    path: list[str]
        The path whose resolution failed.
    """

    def __init__(self, path):
        self.path = path
        super().__init__()


class NoAvailableElementSetsError(Exception):
    """
    No element set is available.
    """


class OutputFileParserNoOutputError(ValueError):
    """
    There was no output for the output file parser to parse.
    """


class SubmissionEnvironmentError(ValueError):
    """
    Raised when submitting a workflow on a machine without a compatible environment.
    """


class MissingEnvironmentExecutableError(SubmissionEnvironmentError):
    """
    The environment does not have the requested executable at all.
    """


class MissingEnvironmentExecutableInstanceError(SubmissionEnvironmentError):
    """
    The environment does not have a suitable instance of the requested executable.
    """


class MissingEnvironmentError(SubmissionEnvironmentError):
    """
    There is no environment with that name.
    """


class UnsupportedScriptDataFormat(ValueError):
    """
    That format of script data is not supported.
    """


class UnknownScriptDataParameter(ValueError):
    """
    Unknown parameter in script data.
    """


class UnknownScriptDataKey(ValueError):
    """
    Unknown key in script data.
    """


class MissingVariableSubstitutionError(KeyError):
    """
    No definition available of a variable being substituted.
    """


class EnvironmentPresetUnknownEnvironmentError(ValueError):
    """
    An environment preset could not be resolved to an execution environment.
    """


class UnknownEnvironmentPresetError(ValueError):
    """
    An execution environment was unknown.
    """


class MultipleEnvironmentsError(ValueError):
    """
    Multiple applicable execution environments exist.
    """


class MissingElementGroup(ValueError):
    """
    An element group should exist but doesn't.
    """



import json


def main_script_test_json_out_obj(p1, _output_files):
    # process
    p1c = {"a": p1 + 100}

    # write outputs
    with _output_files["json"].open("wt") as fh:
        json.dump({"p1c": p1c}, fh)



from PyInstaller.utils.hooks import collect_data_files

from hpcflow.sdk import sdk_classes


# most of the modules in `sdk_classes` are imported on-demand via the app object:
hiddenimports = list(sdk_classes.values())

hiddenimports += [
    "hpcflow.sdk.data",
    "hpcflow.data.demo_data_manifest",
    "hpcflow.data.scripts",
    "hpcflow.data.template_components",
    "hpcflow.data.workflows",
    "hpcflow.tests.data",
    "hpcflow.sdk.core.test_utils",
    "click.testing",
    "requests",  # for GitHub fsspec file system
    "fsspec.implementations.github",  # for GitHub fsspec file system
]

py_include_kwargs = dict(include_py_files=True, excludes=("**/__pycache__",))
datas = (
    collect_data_files("hpcflow.sdk.data")
    + collect_data_files("hpcflow.data.demo_data_manifest")
    + collect_data_files("hpcflow.data.scripts", **py_include_kwargs)
    + collect_data_files("hpcflow.data.template_components")
    + collect_data_files("hpcflow.data.workflows")
    + collect_data_files("hpcflow.tests", **py_include_kwargs)
    + collect_data_files("hpcflow.tests.data")
)



import pytest
from hpcflow.app import app as hf


def pytest_addoption(parser):
    parser.addoption(
        "--slurm",
        action="store_true",
        default=False,
        help="run slurm tests",
    )
    parser.addoption(
        "--wsl",
        action="store_true",
        default=False,
        help="run Windows Subsystem for Linux tests",
    )
    parser.addoption(
        "--direct-linux",
        action="store_true",
        default=False,
        help="run direct-linux submission tests",
    )
    parser.addoption(
        "--integration",
        action="store_true",
        default=False,
        help="run integration-like workflow submission tests",
    )


def pytest_configure(config):
    config.addinivalue_line("markers", "slurm: mark test as slurm to run")
    config.addinivalue_line("markers", "wsl: mark test as wsl to run")
    config.addinivalue_line(
        "markers", "direct_linux: mark test as a direct-linux submission test to run"
    )
    config.addinivalue_line(
        "markers",
        "integration: mark test as an integration-like workflow submission test to run",
    )
    hf.run_time_info.in_pytest = True


def pytest_collection_modifyitems(config, items):
    if config.getoption("--slurm"):
        # --slurm given in cli: only run slurm tests
        for item in items:
            if "slurm" not in item.keywords:
                item.add_marker(pytest.mark.skip(reason="need no --slurm option to run"))
    elif config.getoption("--wsl"):
        # --wsl given in CLI: only run wsl tests
        for item in items:
            if "wsl" not in item.keywords:
                item.add_marker(pytest.mark.skip(reason="need no --wsl option to run"))
    elif config.getoption("--direct-linux"):
        # --direct-linux in CLI: only run these tests
        for item in items:
            if "direct_linux" not in item.keywords:
                item.add_marker(
                    pytest.mark.skip(reason="remove --direct-linux option to run")
                )
    elif config.getoption("--integration"):
        # --integration in CLI: only run these tests
        for item in items:
            if "integration" not in item.keywords:
                item.add_marker(
                    pytest.mark.skip(reason="remove --integration option to run")
                )
    else:
        # --slurm not given in cli: skip slurm tests and do not skip other tests
        for item in items:
            if "slurm" in item.keywords:
                item.add_marker(pytest.mark.skip(reason="need --slurm option to run"))
            elif "wsl" in item.keywords:
                item.add_marker(pytest.mark.skip(reason="need --wsl option to run"))
            elif "direct_linux" in item.keywords:
                item.add_marker(
                    pytest.mark.skip(reason="add --direct_linux option to run")
                )
            elif "integration" in item.keywords:
                item.add_marker(
                    pytest.mark.skip(reason="add --integration option to run")
                )


def pytest_unconfigure(config):
    hf.run_time_info.in_pytest = False


@pytest.fixture
def null_config(tmp_path):
    if not hf.is_config_loaded:
        hf.load_config(config_dir=tmp_path)
    hf.run_time_info.in_pytest = True


@pytest.fixture
def new_null_config(tmp_path):
    hf.load_config(config_dir=tmp_path, warn=False)
    hf.load_template_components(warn=False)
    hf.run_time_info.in_pytest = True



"""
Base model of a shell.
"""

from abc import ABC, abstractmethod
from pathlib import Path
from typing import Dict, List, Optional


class Shell(ABC):
    """Class to represent a shell and templates for jobscript composition.

    This class represents a combination of a shell and an OS. For example, running
    bash on a POSIX OS, and provides snippets that are used to compose a jobscript for
    that combination.

    Parameters
    ----------
    executable: str
        Which executable implements the shell.
    os_args:
        Arguments to pass to the shell.
    """

    def __init__(self, executable=None, os_args=None):
        self._executable = executable or self.DEFAULT_EXE
        self.os_args = os_args

    def __eq__(self, other) -> bool:
        if not isinstance(other, self.__class__):
            return False
        if self._executable == other._executable and self.os_args == other.os_args:
            return True
        return False

    @property
    def executable(self) -> List[str]:
        """
        The executable to use plus any mandatory arguments.
        """
        return [self._executable]

    @property
    def shebang_executable(self) -> str:
        """
        The executable to use in a shebang line.
        """
        return self.executable

    def get_direct_submit_command(self, js_path) -> List[str]:
        """Get the command for submitting a non-scheduled jobscript."""
        return self.executable + [js_path]

    @abstractmethod
    def get_version_info(self, exclude_os: Optional[bool] = False) -> Dict:
        """Get shell and operating system information."""

    def get_wait_command(self, workflow_app_alias: str, sub_idx: int, deps: Dict):
        """
        Get the command to wait for a workflow.
        """
        if deps:
            return (
                f'{workflow_app_alias} workflow $WK_PATH_ARG wait --jobscripts "{sub_idx}:'
                + ",".join(str(i) for i in deps.keys())
                + '"'
            )
        else:
            return ""

    @staticmethod
    def process_app_invoc_executable(app_invoc_exe):
        """
        Perform any post-processing of an application invocation command name.
        """
        return app_invoc_exe

    def process_JS_header_args(self, header_args: Dict) -> Dict:
        """
        Process the application invocation key in the jobscript header arguments.
        """
        app_invoc = self.process_app_invoc_executable(header_args["app_invoc"][0])
        if len(header_args["app_invoc"]) > 1:
            app_invoc += ' "' + header_args["app_invoc"][1] + '"'

        header_args["app_invoc"] = app_invoc
        return header_args

    def prepare_JS_path(self, js_path: Path) -> str:
        """
        Prepare the jobscript path for use.
        """
        return str(js_path)

    def prepare_element_run_dirs(self, run_dirs: List[List[Path]]) -> List[List[str]]:
        """
        Prepare the element run directory names for use.
        """
        return [[str(j) for j in i] for i in run_dirs]



"""
A looping construct for a workflow.
There are multiple types of loop,
notably looping over a set of values or until a condition holds.
"""

from __future__ import annotations

import copy
from typing import Dict, List, Optional, Tuple, Union

from hpcflow.sdk import app
from hpcflow.sdk.core.errors import LoopTaskSubsetError
from hpcflow.sdk.core.json_like import ChildObjectSpec, JSONLike
from hpcflow.sdk.core.loop_cache import LoopCache
from hpcflow.sdk.core.parameters import InputSourceType
from hpcflow.sdk.core.task import WorkflowTask
from hpcflow.sdk.core.utils import check_valid_py_identifier, nth_key, nth_value
from hpcflow.sdk.log import TimeIt

# from .parameters import Parameter

# from valida.conditions import ConditionLike


# @dataclass
# class StoppingCriterion:
#     parameter: Parameter
#     condition: ConditionLike


# @dataclass
# class Loop:
#     parameter: Parameter
#     stopping_criteria: StoppingCriterion
#     # TODO: should be a logical combination of these (maybe provide a superclass in valida to re-use some logic there?)
#     maximum_iterations: int


class Loop(JSONLike):
    """
    A loop in a workflow template.

    Parameters
    ----------
    tasks: list[int | ~hpcflow.app.WorkflowTask]
        List of task insert IDs or workflow tasks.
    num_iterations:
        Number of iterations to perform.
    name: str
        Loop name.
    non_iterable_parameters: list[str]
        Specify input parameters that should not iterate.
    termination: v~hpcflow.app.Rule
        Stopping criterion, expressed as a rule.
    """

    _app_attr = "app"
    _child_objects = (ChildObjectSpec(name="termination", class_name="Rule"),)

    def __init__(
        self,
        tasks: List[Union[int, app.WorkflowTask]],
        num_iterations: int,
        name: Optional[str] = None,
        non_iterable_parameters: Optional[List[str]] = None,
        termination: Optional[app.Rule] = None,
    ) -> None:
        _task_insert_IDs = []
        for task in tasks:
            if isinstance(task, WorkflowTask):
                _task_insert_IDs.append(task.insert_ID)
            elif isinstance(task, int):
                _task_insert_IDs.append(task)
            else:
                raise TypeError(
                    f"`tasks` must be a list whose elements are either task insert IDs "
                    f"or WorkflowTask objects, but received the following: {tasks!r}."
                )

        self._task_insert_IDs = _task_insert_IDs
        self._num_iterations = num_iterations
        self._name = check_valid_py_identifier(name) if name else name
        self._non_iterable_parameters = non_iterable_parameters or []
        self._termination = termination

        self._workflow_template = None  # assigned by parent WorkflowTemplate

    def to_dict(self):
        out = super().to_dict()
        return {k.lstrip("_"): v for k, v in out.items()}

    @classmethod
    def _json_like_constructor(cls, json_like):
        """Invoked by `JSONLike.from_json_like` instead of `__init__`."""
        if "task_insert_IDs" in json_like:
            insert_IDs = json_like.pop("task_insert_IDs")
        else:
            insert_IDs = json_like.pop("tasks")
        obj = cls(tasks=insert_IDs, **json_like)
        return obj

    @property
    def task_insert_IDs(self) -> Tuple[int]:
        """Get the list of task insert_IDs that define the extent of the loop."""
        return tuple(self._task_insert_IDs)

    @property
    def name(self):
        """
        The name of the loop, if one was provided.
        """
        return self._name

    @property
    def num_iterations(self):
        """
        The number of loop iterations to do.
        """
        return self._num_iterations

    @property
    def non_iterable_parameters(self):
        """
        Which parameters are not iterable.
        """
        return self._non_iterable_parameters

    @property
    def termination(self):
        """
        A termination rule for the loop, if one is provided.
        """
        return self._termination

    @property
    def workflow_template(self):
        """
        The workflow template that contains this loop.
        """
        return self._workflow_template

    @workflow_template.setter
    def workflow_template(self, template: app.WorkflowTemplate):
        self._workflow_template = template
        self._validate_against_template()

    @property
    def task_objects(self) -> Tuple[app.WorkflowTask]:
        """
        The tasks in the loop.
        """
        if not self.workflow_template:
            raise RuntimeError(
                "Workflow template must be assigned to retrieve task objects of the loop."
            )
        return tuple(
            self.workflow_template.workflow.tasks.get(insert_ID=i)
            for i in self.task_insert_IDs
        )

    def _validate_against_template(self):
        """Validate the loop parameters against the associated workflow."""

        # insert IDs must exist:
        for insert_ID in self.task_insert_IDs:
            try:
                self.workflow_template.workflow.tasks.get(insert_ID=insert_ID)
            except ValueError:
                raise ValueError(
                    f"Loop {self.name!r} has an invalid task insert ID {insert_ID!r}. "
                    f"Such as task does not exist in the associated workflow."
                )

    def __repr__(self):
        num_iterations_str = ""
        if self.num_iterations is not None:
            num_iterations_str = f", num_iterations={self.num_iterations!r}"

        name_str = ""
        if self.name:
            name_str = f", name={self.name!r}"

        return (
            f"{self.__class__.__name__}("
            f"task_insert_IDs={self.task_insert_IDs!r}{num_iterations_str}{name_str}"
            f")"
        )

    def __deepcopy__(self, memo):
        kwargs = self.to_dict()
        kwargs["tasks"] = kwargs.pop("task_insert_IDs")
        obj = self.__class__(**copy.deepcopy(kwargs, memo))
        obj._workflow_template = self._workflow_template
        return obj


class WorkflowLoop:
    """
    Class to represent a :py:class:`.Loop` that is bound to a
    :py:class:`~hpcflow.app.Workflow`.

    Parameters
    ----------
    index: int
        The index of this loop in the workflow.
    workflow: ~hpcflow.app.Workflow
        The workflow containing this loop.
    template: Loop
        The loop that this was generated from.
    num_added_iterations:
        Description of what iterations have been added.
    iterable_parameters:
        Description of what parameters are being iterated over.
    parents: list[str]
        The paths to the parent entities of this loop.
    """

    _app_attr = "app"

    def __init__(
        self,
        index: int,
        workflow: app.Workflow,
        template: app.Loop,
        num_added_iterations: Dict[Tuple[int], int],
        iterable_parameters: Dict[int : List[int, List[int]]],
        parents: List[str],
    ):
        self._index = index
        self._workflow = workflow
        self._template = template
        self._num_added_iterations = num_added_iterations
        self._iterable_parameters = iterable_parameters
        self._parents = parents

        # appended to on adding a empty loop to the workflow that's a parent of this loop,
        # reset and added to `self._parents` on dump to disk:
        self._pending_parents = []

        # used for `num_added_iterations` when a new loop iteration is added, or when
        # parents are append to; reset to None on dump to disk. Each key is a tuple of
        # parent loop indices and each value is the number of pending new iterations:
        self._pending_num_added_iterations = None

        self._validate()

    @TimeIt.decorator
    def _validate(self):
        # task subset must be a contiguous range of task indices:
        task_indices = self.task_indices
        task_min, task_max = task_indices[0], task_indices[-1]
        if task_indices != tuple(range(task_min, task_max + 1)):
            raise LoopTaskSubsetError(
                f"Loop {self.name!r}: task subset must be an ascending contiguous range, "
                f"but specified task indices were: {self.task_indices!r}."
            )

        for task in self.downstream_tasks:
            for param in self.iterable_parameters:
                if param in task.template.all_schema_input_types:
                    raise NotImplementedError(
                        f"Downstream task {task.unique_name!r} of loop {self.name!r} "
                        f"has as one of its input parameters this loop's iterable "
                        f"parameter {param!r}. This parameter cannot be sourced "
                        f"correctly."
                    )

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(template={self.template!r}, "
            f"num_added_iterations={self.num_added_iterations!r})"
        )

    @property
    def num_added_iterations(self):
        """
        The number of added iterations.
        """
        if self._pending_num_added_iterations:
            return self._pending_num_added_iterations
        else:
            return self._num_added_iterations

    def _initialise_pending_added_iters(self, added_iters_key):
        if not self._pending_num_added_iterations:
            self._pending_num_added_iterations = copy.deepcopy(self._num_added_iterations)

        if added_iters_key not in self._pending_num_added_iterations:
            self._pending_num_added_iterations[added_iters_key] = 1

    def _increment_pending_added_iters(self, added_iters_key):
        if not self._pending_num_added_iterations:
            self._pending_num_added_iterations = copy.deepcopy(self._num_added_iterations)

        self._pending_num_added_iterations[added_iters_key] += 1

    def _update_parents(self, parent: app.WorkflowLoop):
        self._pending_parents.append(parent.name)

        if not self._pending_num_added_iterations:
            self._pending_num_added_iterations = copy.deepcopy(self._num_added_iterations)

        self._pending_num_added_iterations = {
            tuple(list(k) + [0]): v for k, v in self._pending_num_added_iterations.items()
        }

        self.workflow._store.update_loop_parents(
            index=self.index,
            num_added_iters=self.num_added_iterations,
            parents=self.parents,
        )

    def _reset_pending_num_added_iters(self):
        self._pending_num_added_iterations = None

    def _accept_pending_num_added_iters(self):
        if self._pending_num_added_iterations:
            self._num_added_iterations = copy.deepcopy(self._pending_num_added_iterations)
            self._reset_pending_num_added_iters()

    def _reset_pending_parents(self):
        self._pending_parents = []

    def _accept_pending_parents(self):
        self._parents += self._pending_parents
        self._reset_pending_parents()

    @property
    def index(self):
        """
        The index of this loop within its workflow.
        """
        return self._index

    @property
    def task_insert_IDs(self):
        """
        The insertion IDs of the tasks inside this loop.
        """
        return self.template.task_insert_IDs

    @property
    def task_objects(self):
        """
        The tasks in this loop.
        """
        return self.template.task_objects

    @property
    def task_indices(self) -> Tuple[int]:
        """
        The list of task indices that define the extent of the loop.
        """
        return tuple(i.index for i in self.task_objects)

    @property
    def workflow(self):
        """
        The workflow containing this loop.
        """
        return self._workflow

    @property
    def template(self):
        """
        The loop template for this loop.
        """
        return self._template

    @property
    def parents(self) -> List[str]:
        """
        The parents of this loop.
        """
        return self._parents + self._pending_parents

    @property
    def name(self):
        """
        The name of this loop, if one is defined.
        """
        return self.template.name

    @property
    def iterable_parameters(self):
        """
        The parameters that are being iterated over.
        """
        return self._iterable_parameters

    @property
    def num_iterations(self):
        """
        The number of iterations.
        """
        return self.template.num_iterations

    @property
    def downstream_tasks(self) -> List[app.WorkflowLoop]:
        """Tasks that are not part of the loop, and downstream from this loop."""
        return self.workflow.tasks[self.task_objects[-1].index + 1 :]

    @property
    def upstream_tasks(self) -> List[app.WorkflowLoop]:
        """Tasks that are not part of the loop, and upstream from this loop."""
        return self.workflow.tasks[: self.task_objects[0].index]

    @staticmethod
    @TimeIt.decorator
    def _find_iterable_parameters(loop_template: app.Loop):
        all_inputs_first_idx = {}
        all_outputs_idx = {}
        for task in loop_template.task_objects:
            for typ in task.template.all_schema_input_types:
                if typ not in all_inputs_first_idx:
                    all_inputs_first_idx[typ] = task.insert_ID
            for typ in task.template.all_schema_output_types:
                if typ not in all_outputs_idx:
                    all_outputs_idx[typ] = []
                all_outputs_idx[typ].append(task.insert_ID)

        iterable_params = {}
        for typ, first_idx in all_inputs_first_idx.items():
            if typ in all_outputs_idx and first_idx <= all_outputs_idx[typ][0]:
                iterable_params[typ] = {
                    "input_task": first_idx,
                    "output_tasks": all_outputs_idx[typ],
                }

        for non_iter in loop_template.non_iterable_parameters:
            if non_iter in iterable_params:
                del iterable_params[non_iter]

        return iterable_params

    @classmethod
    @TimeIt.decorator
    def new_empty_loop(
        cls,
        index: int,
        workflow: app.Workflow,
        template: app.Loop,
        iter_loop_idx: List[Dict],
    ) -> Tuple[app.WorkflowLoop, List[Dict[str, int]]]:
        """
        Make a new empty loop.

        Parameters
        ----------
        index: int
            The index of the loop to create.
        workflow: ~hpcflow.app.Workflow
            The workflow that will contain the loop.
        template: Loop
            The template for the loop.
        iter_loop_idx: list[dict]
            Iteration information from parent loops.
        """
        parent_loops = cls._get_parent_loops(index, workflow, template)
        parent_names = [i.name for i in parent_loops]
        num_added_iters = {}
        for i in iter_loop_idx:
            num_added_iters[tuple([i[j] for j in parent_names])] = 1

        obj = cls(
            index=index,
            workflow=workflow,
            template=template,
            num_added_iterations=num_added_iters,
            iterable_parameters=cls._find_iterable_parameters(template),
            parents=parent_names,
        )
        return obj

    @classmethod
    @TimeIt.decorator
    def _get_parent_loops(
        cls,
        index: int,
        workflow: app.Workflow,
        template: app.Loop,
    ) -> List[app.WorkflowLoop]:
        parents = []
        passed_self = False
        self_tasks = set(template.task_insert_IDs)
        for loop_i in workflow.loops:
            if loop_i.index == index:
                passed_self = True
                continue
            other_tasks = set(loop_i.task_insert_IDs)
            if self_tasks.issubset(other_tasks):
                if (self_tasks == other_tasks) and not passed_self:
                    continue
                parents.append(loop_i)
        return parents

    @TimeIt.decorator
    def get_parent_loops(self) -> List[app.WorkflowLoop]:
        """Get loops whose task subset is a superset of this loop's task subset. If two
        loops have identical task subsets, the first loop in the workflow loop list is
        considered the child."""
        return self._get_parent_loops(self.index, self.workflow, self.template)

    @TimeIt.decorator
    def get_child_loops(self) -> List[app.WorkflowLoop]:
        """Get loops whose task subset is a subset of this loop's task subset. If two
        loops have identical task subsets, the first loop in the workflow loop list is
        considered the child."""
        children = []
        passed_self = False
        self_tasks = set(self.task_insert_IDs)
        for loop_i in self.workflow.loops:
            if loop_i.index == self.index:
                passed_self = True
                continue
            other_tasks = set(loop_i.task_insert_IDs)
            if self_tasks.issuperset(other_tasks):
                if (self_tasks == other_tasks) and passed_self:
                    continue
                children.append(loop_i)

        # order by depth, so direct child is first:
        children = sorted(children, key=lambda x: len(next(iter(x.num_added_iterations))))
        return children

    @TimeIt.decorator
    def add_iteration(self, parent_loop_indices=None, cache: Optional[LoopCache] = None):
        """
        Add an iteration to this loop.

        Parameters
        ----------
        parent_loop_indices:
            Where have any parent loops got up to?
        cache:
            A cache used to make adding the iteration more efficient.
            One will be created if it is not supplied.
        """
        if not cache:
            cache = LoopCache.build(self.workflow)
        parent_loops = self.get_parent_loops()
        child_loops = self.get_child_loops()
        parent_loop_indices = parent_loop_indices or {}
        if parent_loops and not parent_loop_indices:
            parent_loop_indices = {i.name: 0 for i in parent_loops}

        iters_key = tuple([parent_loop_indices[k] for k in self.parents])
        cur_loop_idx = self.num_added_iterations[iters_key] - 1
        all_new_data_idx = {}  # keys are (task.insert_ID and element.index)

        # initialise a new `num_added_iterations` key on each child loop:
        for child in child_loops:
            iters_key_dct = {
                **parent_loop_indices,
                self.name: cur_loop_idx + 1,
            }
            added_iters_key_chd = tuple([iters_key_dct.get(j, 0) for j in child.parents])
            child._initialise_pending_added_iters(added_iters_key_chd)

        for task in self.task_objects:

            new_loop_idx = {
                **parent_loop_indices,
                self.name: cur_loop_idx + 1,
                **{
                    child.name: 0
                    for child in child_loops
                    if task.insert_ID in child.task_insert_IDs
                },
            }
            added_iter_IDs = []
            for elem_idx in range(task.num_elements):

                elem_ID = task.element_IDs[elem_idx]

                new_data_idx = {}

                # copy resources from zeroth iteration:
                zeroth_iter_ID, zi_iter_data_idx = cache.zeroth_iters[elem_ID]
                zi_elem_ID, zi_idx = cache.iterations[zeroth_iter_ID]
                zi_data_idx = nth_value(cache.data_idx[zi_elem_ID], zi_idx)

                for key, val in zi_data_idx.items():
                    if key.startswith("resources."):
                        new_data_idx[key] = val

                for inp in task.template.all_schema_inputs:
                    is_inp_task = False
                    iter_dat = self.iterable_parameters.get(inp.typ)
                    if iter_dat:
                        is_inp_task = task.insert_ID == iter_dat["input_task"]

                    if is_inp_task:
                        # source from final output task of previous iteration, with all parent
                        # loop indices the same as previous iteration, and all child loop indices
                        # maximised:

                        # identify element(s) from which this iterable input should be
                        # parametrised:
                        if task.insert_ID == iter_dat["output_tasks"][-1]:
                            src_elem_ID = elem_ID
                            grouped_elems = None
                        else:
                            src_elem_IDs_all = cache.element_dependents[elem_ID]
                            src_elem_IDs = {
                                k: v
                                for k, v in src_elem_IDs_all.items()
                                if cache.elements[k]["task_insert_ID"]
                                == iter_dat["output_tasks"][-1]
                            }
                            # consider groups
                            inp_group_name = inp.single_labelled_data.get("group")
                            grouped_elems = []
                            for src_elem_j_ID, src_elem_j_dat in src_elem_IDs.items():
                                i_in_group = any(
                                    k == inp_group_name
                                    for k in src_elem_j_dat["group_names"]
                                )
                                if i_in_group:
                                    grouped_elems.append(src_elem_j_ID)

                            if not grouped_elems and len(src_elem_IDs) > 1:
                                raise NotImplementedError(
                                    f"Multiple elements found in the iterable parameter "
                                    f"{inp!r}'s latest output task (insert ID: "
                                    f"{iter_dat['output_tasks'][-1]}) that can be used "
                                    f"to parametrise the next iteration: "
                                    f"{list(src_elem_IDs.keys())!r}."
                                )

                            elif not src_elem_IDs:
                                # TODO: maybe OK?
                                raise NotImplementedError(
                                    f"No elements found in the iterable parameter "
                                    f"{inp!r}'s latest output task (insert ID: "
                                    f"{iter_dat['output_tasks'][-1]}) that can be used "
                                    f"to parametrise the next iteration."
                                )

                            else:
                                src_elem_ID = nth_key(src_elem_IDs, 0)

                        child_loop_max_iters = {}
                        parent_loop_same_iters = {
                            i.name: parent_loop_indices[i.name] for i in parent_loops
                        }
                        child_iter_parents = {
                            **parent_loop_same_iters,
                            self.name: cur_loop_idx,
                        }
                        for i in child_loops:
                            i_num_iters = i.num_added_iterations[
                                tuple(child_iter_parents[j] for j in i.parents)
                            ]
                            i_max = i_num_iters - 1
                            child_iter_parents[i.name] = i_max
                            child_loop_max_iters[i.name] = i_max

                        source_iter_loop_idx = {
                            **child_loop_max_iters,
                            **parent_loop_same_iters,
                            self.name: cur_loop_idx,
                        }

                        # identify the ElementIteration from which this input should be
                        # parametrised:
                        loop_idx_key = tuple(sorted(source_iter_loop_idx.items()))
                        if grouped_elems:
                            src_data_idx = []
                            for src_elem_ID in grouped_elems:
                                src_data_idx.append(
                                    cache.data_idx[src_elem_ID][loop_idx_key]
                                )
                        else:
                            src_data_idx = cache.data_idx[src_elem_ID][loop_idx_key]

                        if not src_data_idx:
                            raise RuntimeError(
                                f"Could not find a source iteration with loop_idx: "
                                f"{source_iter_loop_idx!r}."
                            )

                        if grouped_elems:
                            inp_dat_idx = [i[f"outputs.{inp.typ}"] for i in src_data_idx]
                        else:
                            inp_dat_idx = src_data_idx[f"outputs.{inp.typ}"]
                        new_data_idx[f"inputs.{inp.typ}"] = inp_dat_idx

                    else:
                        inp_key = f"inputs.{inp.typ}"

                        orig_inp_src = cache.elements[elem_ID]["input_sources"][inp_key]
                        inp_dat_idx = None

                        if orig_inp_src.source_type is InputSourceType.LOCAL:
                            # keep locally defined inputs from original element
                            inp_dat_idx = zi_data_idx[inp_key]

                        elif orig_inp_src.source_type is InputSourceType.DEFAULT:
                            # keep default value from original element
                            try:
                                inp_dat_idx = zi_data_idx[inp_key]
                            except KeyError:
                                # if this input is required by a conditional action, and
                                # that condition is not met, then this input will not
                                # exist in the action-run data index, so use the initial
                                # iteration data index:
                                inp_dat_idx = zi_iter_data_idx[inp_key]

                        elif orig_inp_src.source_type is InputSourceType.TASK:
                            if orig_inp_src.task_ref not in self.task_insert_IDs:
                                # source the data_idx from the iteration with same parent
                                # loop indices as the new iteration to add:
                                # src_iters = []
                                src_data_idx = []
                                for li_k, di_k in cache.data_idx[elem_ID].items():
                                    skip_iter = False
                                    li_k_dct = dict(li_k)
                                    for p_k, p_v in parent_loop_indices.items():
                                        if li_k_dct.get(p_k) != p_v:
                                            skip_iter = True
                                            break
                                    if not skip_iter:
                                        src_data_idx.append(di_k)

                                # could be multiple, but they should all have the same
                                # data index for this parameter:
                                src_data_idx = src_data_idx[0]
                                inp_dat_idx = src_data_idx[inp_key]
                            else:
                                is_group = False
                                if (
                                    not inp.multiple
                                    and "group" in inp.single_labelled_data
                                ):
                                    # this input is a group, assume for now all elements:
                                    is_group = True

                                # same task/element, but update iteration to the just-added
                                # iteration:
                                key_prefix = orig_inp_src.task_source_type.name.lower()
                                prev_dat_idx_key = f"{key_prefix}s.{inp.typ}"
                                new_sources = []
                                for (
                                    tiID,
                                    e_idx,
                                ), prev_dat_idx in all_new_data_idx.items():
                                    if tiID == orig_inp_src.task_ref:
                                        # find which element in that task `element`
                                        # depends on:
                                        task_i = self.workflow.tasks.get(insert_ID=tiID)
                                        elem_i_ID = task_i.element_IDs[e_idx]
                                        src_elem_IDs_all = cache.element_dependents[
                                            elem_i_ID
                                        ]
                                        src_elem_IDs_i = {
                                            k: v
                                            for k, v in src_elem_IDs_all.items()
                                            if cache.elements[k]["task_insert_ID"]
                                            == task.insert_ID
                                        }

                                        # filter src_elem_IDs_i for matching element IDs:
                                        src_elem_IDs_i = [
                                            i for i in src_elem_IDs_i if i == elem_ID
                                        ]
                                        if (
                                            len(src_elem_IDs_i) == 1
                                            and src_elem_IDs_i[0] == elem_ID
                                        ):
                                            new_sources.append((tiID, e_idx))

                                if is_group:
                                    inp_dat_idx = [
                                        all_new_data_idx[i][prev_dat_idx_key]
                                        for i in new_sources
                                    ]
                                else:
                                    assert len(new_sources) == 1
                                    prev_dat_idx = all_new_data_idx[new_sources[0]]
                                    inp_dat_idx = prev_dat_idx[prev_dat_idx_key]

                        if inp_dat_idx is None:
                            raise RuntimeError(
                                f"Could not find a source for parameter {inp.typ} "
                                f"when adding a new iteration for task {task!r}."
                            )

                        new_data_idx[inp_key] = inp_dat_idx

                # add any locally defined sub-parameters:
                inp_statuses = cache.elements[elem_ID]["input_statuses"]
                inp_status_inps = set([f"inputs.{i}" for i in inp_statuses])
                sub_params = inp_status_inps - set(new_data_idx.keys())
                for sub_param_i in sub_params:
                    sub_param_data_idx_iter_0 = zi_data_idx
                    try:
                        sub_param_data_idx = sub_param_data_idx_iter_0[sub_param_i]
                    except KeyError:
                        # as before, if this input is required by a conditional action,
                        # and that condition is not met, then this input will not exist in
                        # the action-run data index, so use the initial iteration data
                        # index:
                        sub_param_data_idx = zi_data_idx[sub_param_i]

                    new_data_idx[sub_param_i] = sub_param_data_idx

                for out in task.template.all_schema_outputs:
                    path_i = f"outputs.{out.typ}"
                    p_src = {"type": "EAR_output"}
                    new_data_idx[path_i] = self.workflow._add_unset_parameter_data(p_src)

                schema_params = set(
                    i for i in new_data_idx.keys() if len(i.split(".")) == 2
                )
                all_new_data_idx[(task.insert_ID, elem_idx)] = new_data_idx

                iter_ID_i = self.workflow._store.add_element_iteration(
                    element_ID=elem_ID,
                    data_idx=new_data_idx,
                    schema_parameters=list(schema_params),
                    loop_idx=new_loop_idx,
                )
                if cache:
                    cache.add_iteration(
                        iter_ID=iter_ID_i,
                        task_insert_ID=task.insert_ID,
                        element_ID=elem_ID,
                        loop_idx=new_loop_idx,
                        data_idx=new_data_idx,
                    )

                added_iter_IDs.append(iter_ID_i)

            task.initialise_EARs(iter_IDs=added_iter_IDs)

        added_iters_key = tuple(parent_loop_indices[k] for k in self.parents)
        self._increment_pending_added_iters(added_iters_key)
        self.workflow._store.update_loop_num_iters(
            index=self.index,
            num_added_iters=self.num_added_iterations,
        )

        # add iterations to fixed-number-iteration children only:
        for child in child_loops[::-1]:
            if child.num_iterations is not None:
                for _ in range(child.num_iterations - 1):
                    par_idx = {k: 0 for k in child.parents}
                    child.add_iteration(
                        parent_loop_indices={
                            **par_idx,
                            **parent_loop_indices,
                            self.name: cur_loop_idx + 1,
                        },
                        cache=cache,
                    )

    def test_termination(self, element_iter):
        """Check if a loop should terminate, given the specified completed element
        iteration."""
        if self.template.termination:
            return self.template.termination.test(element_iter)
        return False
